import math
from dataclasses import dataclass
from typing import Union, Callable, Type, Dict, Literal, Tuple, Any

from aislib.misc_utils import get_logger
from torch import nn
from transformers import PreTrainedModel, AutoModel, AutoConfig

from eir.models.models_base import get_output_dimensions_for_input
from eir.models.sequence.transformer_models import (
    TransformerFeatureExtractor,
    PerceiverIOFeatureExtractor,
    SequenceModelConfig,
    TransformerWrapperModel,
    BasicTransformerFeatureExtractorModelConfig,
    PerceiverIOModelConfig,
    get_embedding_dim_for_sequence_model,
)
from eir.setup.setup_utils import get_unsupported_hf_models

logger = get_logger(name=__name__)


@dataclass
class SequenceModelObjectsForWrapperModel:
    feature_extractor: Union[
        TransformerFeatureExtractor, PerceiverIOFeatureExtractor, nn.Module
    ]
    embeddings: Union[None, nn.Module]
    embedding_dim: int
    external: bool
    known_out_features: Union[None, int]


def get_sequence_model(
    sequence_model_config: SequenceModelConfig,
    model_registry_lookup: Callable[[str], Type[nn.Module]],
    num_tokens: int,
    max_length: int,
    embedding_dim: int,
    device: str,
) -> TransformerWrapperModel:
    feature_extractor_max_length = max_length
    num_chunks = 1
    if sequence_model_config.window_size:
        logger.info(
            "Using sliding model for sequence input as window size was set to %d.",
            sequence_model_config.window_size,
        )
        feature_extractor_max_length = sequence_model_config.window_size
        num_chunks = math.ceil(max_length / sequence_model_config.window_size)

    objects_for_wrapper = _get_sequence_feature_extractor_objects_for_wrapper_model(
        model_type=sequence_model_config.model_type,
        model_registry_lookup=model_registry_lookup,
        pretrained=sequence_model_config.pretrained_model,
        pretrained_frozen=sequence_model_config.freeze_pretrained_model,
        model_config=sequence_model_config.model_init_config,
        num_tokens=num_tokens,
        embedding_dim=embedding_dim,
        feature_extractor_max_length=feature_extractor_max_length,
        num_chunks=num_chunks,
        pool=sequence_model_config.pool,
    )

    wrapper_model_class = model_registry_lookup(model_type="sequence-wrapper-default")
    sequence_model = wrapper_model_class(
        feature_extractor=objects_for_wrapper.feature_extractor,
        external_feature_extractor=objects_for_wrapper.external,
        model_config=sequence_model_config,
        embedding_dim=objects_for_wrapper.embedding_dim,
        num_tokens=num_tokens,
        max_length=max_length,
        embeddings=objects_for_wrapper.embeddings,
        device=device,
        pre_computed_num_out_features=objects_for_wrapper.known_out_features,
    ).to(device=device)

    return sequence_model


def _get_sequence_feature_extractor_objects_for_wrapper_model(
    model_type: str,
    model_registry_lookup: Callable[[str], Type[nn.Module]],
    pretrained: bool,
    pretrained_frozen: bool,
    model_config: Union[
        BasicTransformerFeatureExtractorModelConfig, PerceiverIOModelConfig, Dict
    ],
    num_tokens: int,
    embedding_dim: int,
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    if "sequence-default" in model_type or model_type.startswith("eir-"):
        model_class = model_registry_lookup(model_type=model_type)
        objects_for_wrapper = _get_basic_sequence_feature_extractor_objects(
            model_config=model_config,
            num_tokens=num_tokens,
            feature_extractor_max_length=feature_extractor_max_length,
            embedding_dim=embedding_dim,
            feature_extractor_class=model_class,
        )
    elif model_type == "perceiver":
        objects_for_wrapper = _get_perceiver_sequence_feature_extractor_objects(
            model_config=model_config,
            max_length=feature_extractor_max_length,
            num_chunks=num_chunks,
        )
    elif pretrained:
        objects_for_wrapper = _get_pretrained_hf_sequence_feature_extractor_objects(
            model_name=model_type,
            frozen=pretrained_frozen,
            feature_extractor_max_length=feature_extractor_max_length,
            num_chunks=num_chunks,
            num_tokens=num_tokens,
            pool=pool,
        )
    else:
        objects_for_wrapper = _get_hf_sequence_feature_extractor_objects(
            model_name=model_type,
            model_config=model_config,
            feature_extractor_max_length=feature_extractor_max_length,
            num_chunks=num_chunks,
            pool=pool,
        )

    return objects_for_wrapper


def _get_manual_out_features_for_external_feature_extractor(
    input_length: int,
    embedding_dim: int,
    num_chunks: int,
    feature_extractor: nn.Module,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> int:
    input_shape = _get_sequence_input_dim(
        input_length=input_length,
        embedding_dim=embedding_dim,
    )
    out_feature_shape = get_output_dimensions_for_input(
        module=feature_extractor,
        input_shape=input_shape,
        hf_model=True,
        pool=pool,
    )
    manual_out_features = out_feature_shape.numel() * num_chunks

    return manual_out_features


def _get_sequence_input_dim(
    input_length: int, embedding_dim: int
) -> Tuple[int, int, int]:
    return 1, input_length, embedding_dim


def _get_pretrained_hf_sequence_feature_extractor_objects(
    model_name: str,
    num_tokens: int,
    frozen: bool,
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    _warn_about_unsupported_hf_model(model_name=model_name)

    pretrained_model = _get_hf_pretrained_model(model_name=model_name)
    pretrained_model.resize_token_embeddings(new_num_tokens=num_tokens)
    pretrained_model_embeddings = pretrained_model.get_input_embeddings()
    feature_extractor = pretrained_model

    if frozen:
        logger.info("Freezing weights and embeddings of model '%s'.", model_name)
        for param in feature_extractor.parameters():
            param.requires_grad = False
        for param in pretrained_model_embeddings.parameters():
            param.requires_grad = False

    pretrained_embedding_dim = _pretrained_hf_model_embedding_dim(
        embeddings=pretrained_model_embeddings,
        model_name=model_name,
    )
    known_out_features = _get_manual_out_features_for_external_feature_extractor(
        input_length=feature_extractor_max_length,
        embedding_dim=pretrained_embedding_dim,
        num_chunks=num_chunks,
        feature_extractor=feature_extractor,
        pool=pool,
    )
    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=pretrained_model,
        embeddings=pretrained_model_embeddings,
        embedding_dim=pretrained_embedding_dim,
        external=True,
        known_out_features=known_out_features,
    )

    return objects_for_wrapper


def _get_hf_pretrained_model(model_name: str) -> PreTrainedModel:
    pretrained_model = AutoModel.from_pretrained(
        pretrained_model_name_or_path=model_name
    )
    logger.info(
        "Loaded external pre-trained model '%s'. Note that this means that "
        "many configurations that might be set in input_type_info and model_config"
        "(e.g. 'embedding_dim') have no effect, as the default settings for "
        "the pre-trained model are used.",
        model_name,
    )
    return pretrained_model


def _get_hf_sequence_feature_extractor_objects(
    model_name: str,
    model_config: Dict[str, Any],
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    _warn_about_unsupported_hf_model(model_name=model_name)

    feature_extractor = _get_hf_model(model_name=model_name, model_config=model_config)
    pretrained_model_embeddings = feature_extractor.get_input_embeddings()

    pretrained_embedding_dim = _pretrained_hf_model_embedding_dim(
        embeddings=pretrained_model_embeddings,
        model_name=model_name,
    )
    known_out_features = _get_manual_out_features_for_external_feature_extractor(
        input_length=feature_extractor_max_length,
        embedding_dim=pretrained_embedding_dim,
        num_chunks=num_chunks,
        feature_extractor=feature_extractor,
        pool=pool,
    )
    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=feature_extractor,
        embeddings=None,
        embedding_dim=pretrained_embedding_dim,
        external=True,
        known_out_features=known_out_features,
    )

    return objects_for_wrapper


def _pretrained_hf_model_embedding_dim(
    embeddings: nn.Module, model_name: str = ""
) -> int:
    if hasattr(embeddings, "embedding_dim"):
        return embeddings.embedding_dim
    elif hasattr(embeddings, "dim"):
        return embeddings.dim
    elif hasattr(embeddings, "emb_layers"):
        return embeddings.emb_layers[0].embedding_dim

    raise ValueError(f"Could not find embedding dimension for model {model_name}.")


def _get_hf_model(model_name: str, model_config: Dict[str, Any]) -> nn.Module:
    config = AutoConfig.for_model(model_type=model_name, **model_config)
    model = AutoModel.from_config(config=config)
    logger.info(
        "Set up external (not using pre-trained weights) model '%s'. "
        "With configuration %s. Note that setting up external models ignores values "
        "for fields 'embedding_dim' and 'max_length' from input_type_info "
        "configuration. To configure these models, set the relevant values in the "
        "model_config field of the input configuration.",
        model_name,
        config,
    )
    return model


def _get_basic_sequence_feature_extractor_objects(
    model_config: BasicTransformerFeatureExtractorModelConfig,
    num_tokens: int,
    feature_extractor_max_length: int,
    embedding_dim: int,
    feature_extractor_class: Callable = TransformerFeatureExtractor,
) -> SequenceModelObjectsForWrapperModel:
    parsed_embedding_dim = get_embedding_dim_for_sequence_model(
        embedding_dim=embedding_dim,
        num_tokens=num_tokens,
        num_heads=model_config.num_heads,
    )

    feature_extractor = feature_extractor_class(
        model_config=model_config,
        num_tokens=num_tokens,
        max_length=feature_extractor_max_length,
        embedding_dim=parsed_embedding_dim,
    )

    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=feature_extractor,
        embeddings=None,
        embedding_dim=parsed_embedding_dim,
        external=False,
        known_out_features=0,
    )

    return objects_for_wrapper


def _get_perceiver_sequence_feature_extractor_objects(
    model_config: PerceiverIOModelConfig,
    max_length: int,
    num_chunks: int,
):
    feature_extractor = PerceiverIOFeatureExtractor(
        model_config=model_config,
        max_length=max_length,
    )

    known_out_features = feature_extractor.num_out_features * num_chunks

    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=feature_extractor,
        embeddings=None,
        embedding_dim=model_config.dim,
        external=False,
        known_out_features=known_out_features,
    )

    return objects_for_wrapper


def _sequence_model_registry(model_type: str) -> Type[nn.Module]:
    if model_type == "sequence-default":
        return TransformerFeatureExtractor
    elif model_type == "sequence-wrapper-default":
        return TransformerWrapperModel
    else:
        raise ValueError()


def _warn_about_unsupported_hf_model(model_name: str) -> None:
    unsupported_models = get_unsupported_hf_models()
    if model_name in unsupported_models.keys():
        reason = unsupported_models[model_name]
        logger.warning(
            "Model '%s' has not been tested for compatibility with EIR due to "
            "reason: '%s'. It is very likely that it will not work straight out of "
            "the box with EIR.",
            model_name,
            reason,
        )



from dataclasses import dataclass
from typing import Callable, TYPE_CHECKING, Literal

import torch
from torch import nn

from eir.models.omics.models_locally_connected import flatten_h_w_fortran

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions


@dataclass
class IdentityModelConfig:
    """
    :param flatten:
        Whether to flatten the input.

    :param flatten_shape:
        What column-row order to flatten the input in.
    """

    flatten: bool = True
    flatten_shape: Literal["c", "fortran"] = "c"


class IdentityModel(nn.Module):
    def __init__(
        self, model_config: IdentityModelConfig, data_dimensions: "DataDimensions"
    ):
        super().__init__()

        self.model_config = model_config
        self.data_dimensions = data_dimensions

        self.process_func = get_identity_reshape_func(
            flatten=self.model_config.flatten,
            flatten_shape=self.model_config.flatten_shape,
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.data_dimensions.num_elements()

    @property
    def num_out_features(self) -> int:
        return self.fc_1_in_features

    def _init_weights(self):
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = self.process_func(input)
        return out


def get_identity_reshape_func(
    flatten: bool, flatten_shape: Literal["c", "fortran"]
) -> Callable[[torch.Tensor], torch.Tensor]:
    if not flatten:
        return lambda x: x

    if flatten_shape == "fortran":
        return flatten_h_w_fortran
    else:
        return lambda x: x.view(x.shape[0], -1)



from typing import Sequence, Dict, Tuple

import colorama as clr


def log_and_raise_missing_or_multiple_input_matches(
    train_name: str,
    train_type: str,
    matches: Sequence[Dict],
    predict_names_and_types: Sequence[Tuple[str, str]],
) -> None:
    train_name = clr.Style.BRIGHT + clr.Fore.GREEN + train_name + clr.Style.RESET_ALL
    train_type = clr.Style.BRIGHT + clr.Fore.GREEN + train_type + clr.Style.RESET_ALL

    predict_names_and_types_formatted = ""
    for name, type_ in predict_names_and_types:
        name = clr.Style.BRIGHT + clr.Fore.RED + name + clr.Style.RESET_ALL
        type_ = clr.Style.BRIGHT + clr.Fore.RED + type_ + clr.Style.RESET_ALL
        predict_names_and_types_formatted += f"('{name}', '{type_}'), "
    predict_names_and_types_formatted = predict_names_and_types_formatted[:-2]

    if len(matches) == 0:
        msg = (
            f"When running the predict module, no *predict* input config was found "
            f"that matches a configuration used during training. There must be "
            f"exactly one match for each input config used during training. "
            f"The following input configs were used during training and no match "
            f"was found (name, type): ('{train_name}', '{train_type}'). "
            f"The following input configs were found in the *predict* config file(s): "
            f"{predict_names_and_types_formatted}. "
            f"To fix this, make sure that the input configs passed to the predict "
            f"module match the input configs used during training."
        )
    else:
        msg = (
            f"When running the predict module, multiple *predict* input configs were "
            f"found that match a configuration used during training. There must be "
            f"exactly one match for each input config used during training. "
            f"The following input configs were used during training and multiple "
            f"was found (name, type): ('{train_name}', '{train_type}'). "
            f"The following input configs were found in the *predict* config file(s): "
            f"{predict_names_and_types_formatted}. "
            f"To fix this, make sure that the input configs passed to the predict "
            f"module match the input configs used during training."
        )

    msg = msg.replace(
        "*predict*", clr.Style.BRIGHT + clr.Fore.RED + "*predict*" + clr.Style.RESET_ALL
    )

    raise ValueError(msg)


def log_and_raise_missing_or_multiple_output_matches(
    train_name: str,
    train_type: str,
    train_cat_columns: Sequence[str],
    train_con_columns: Sequence[str],
    matches: Sequence[Dict],
    predict_names_and_types: Sequence[Tuple[str, str, Sequence[str], Sequence[str]]],
) -> None:
    train_name = clr.Style.BRIGHT + clr.Fore.GREEN + train_name + clr.Style.RESET_ALL
    train_type = clr.Style.BRIGHT + clr.Fore.GREEN + train_type + clr.Style.RESET_ALL
    train_cat_columns = (
        clr.Style.BRIGHT + clr.Fore.GREEN + str(train_cat_columns) + clr.Style.RESET_ALL
    )
    train_con_columns = (
        clr.Style.BRIGHT + clr.Fore.GREEN + str(train_con_columns) + clr.Style.RESET_ALL
    )

    predict_names_and_types_formatted = ""
    for name, type_, cat_cols, con_cols in predict_names_and_types:
        name = clr.Style.BRIGHT + clr.Fore.RED + name + clr.Style.RESET_ALL
        type_ = clr.Style.BRIGHT + clr.Fore.RED + type_ + clr.Style.RESET_ALL
        cat_cols = clr.Style.BRIGHT + clr.Fore.RED + str(cat_cols) + clr.Style.RESET_ALL
        con_cols = clr.Style.BRIGHT + clr.Fore.RED + str(con_cols) + clr.Style.RESET_ALL

        predict_names_and_types_formatted += (
            f"('{name}', '{type_}', cat_cols: '{cat_cols}', con_cols: '{con_cols}'), "
        )
    predict_names_and_types_formatted = predict_names_and_types_formatted[:-2]

    if len(matches) == 0:
        msg = (
            f"When running the predict module, no *predict* output config was found "
            f"that matches a configuration used during training. There must be "
            f"exactly one match for each output config used during training. "
            f"The following output configs were used during training and no match "
            f"was found (name, type): ('{train_name}', '{train_type}') "
            f"with target columns "
            f"cat: {train_cat_columns} and con: {train_con_columns}. "
            f"The following output configs were found in the *predict* config file(s): "
            f"{predict_names_and_types_formatted}. "
            f"To fix this, make sure that the output configs passed to the predict "
            f"module match the output configs used during training."
        )
    else:
        msg = (
            f"When running the predict module, multiple *predict* output configs were "
            f"found that match a configuration used during training. There must be "
            f"exactly one match for each output config used during training. "
            f"The following output configs were used during training and multiple "
            f"was found (name, type): ('{train_name}', '{train_type}') "
            f"with target columns "
            f"cat: {train_cat_columns} and con: {train_con_columns}. "
            f"The following output configs were found in the *predict* config file(s): "
            f"{predict_names_and_types_formatted}. "
            f"To fix this, make sure that the output configs passed to the predict "
            f"module match the output configs used during training."
        )

    msg = msg.replace(
        "*predict*", clr.Style.BRIGHT + clr.Fore.RED + "*predict*" + clr.Style.RESET_ALL
    )
    raise ValueError(msg)



from typing import Iterable, Sequence

import timm
import torch
from torch_optimizer import _NAME_OPTIM_MAP
from tqdm import tqdm
from transformers.models.auto.modeling_auto import MODEL_MAPPING_NAMES


class RunningStatistics:
    """
    Adapted from https://gist.github.com/thomasbrandon/ad5b1218fc573c10ea4e1f0c63658469.
    """

    def __init__(self, n_dims: int = 2):
        self.n_dims = n_dims
        self.n = 0
        self.sum = 0

        self.shape = None
        self.num_var = None

    def update(self, data: torch.Tensor):
        data = data.view(*list(data.shape[: -self.n_dims]) + [-1])

        with torch.no_grad():
            new_n = data.shape[-1]
            new_var = data.var(-1)
            new_sum = data.sum(-1)

            if self.n == 0:
                self.n = new_n
                self.shape = data.shape[:-1]
                self.sum = new_sum
                self.num_var = new_var.mul_(new_n)

            else:
                ratio = self.n / new_n
                t = (self.sum / ratio).sub_(new_sum).pow_(2)

                self.num_var.add_(other=new_var, alpha=new_n)
                self.num_var.add_(other=t, alpha=ratio / (self.n + new_n))

                self.sum += new_sum
                self.n += new_n

    @property
    def mean(self):
        return self.sum / self.n if self.n > 0 else None

    @property
    def var(self):
        return self.num_var / self.n if self.n > 0 else None

    @property
    def std(self):
        return self.var.sqrt() if self.n > 0 else None


def collect_stats(
    tensor_iterable: Iterable[torch.Tensor], n_dims: int = 2
) -> RunningStatistics:
    stats = RunningStatistics(n_dims=n_dims)
    for it in tqdm(tensor_iterable, desc="Gathering Statistics: "):
        if hasattr(it, "data"):
            stats.update(it.data)
        else:
            stats.update(it)
    return stats


def get_base_optimizer_names() -> set:
    base_names = {"sgdm", "adam", "adamw", "adahessian", "adabelief", "adabeliefw"}

    return base_names


def get_all_optimizer_names() -> Sequence[str]:
    external_optimizers = set(_NAME_OPTIM_MAP.keys())
    base_optimizers = get_base_optimizer_names()
    all_optimizers = set.union(base_optimizers, external_optimizers)
    all_optimizers = sorted(list(all_optimizers))

    return all_optimizers


def get_all_timm_model_names() -> Sequence[str]:
    pretrained_names = {i for i in timm.list_models() if not i.startswith("tf")}
    other_model_classes = {i for i in dir(timm.models) if "Net" in i}
    all_models = set.union(pretrained_names, other_model_classes)
    all_models_list = sorted(list(all_models))

    return all_models_list


def get_all_hf_model_names() -> Sequence[str]:
    all_models = sorted(list(MODEL_MAPPING_NAMES.keys()))
    unsupported = get_unsupported_hf_models()
    unsupported_names = unsupported.keys()
    return [i for i in all_models if i not in unsupported_names]


def get_unsupported_hf_models() -> dict:
    unsupported = {
        "audio-spectrogram-transformer": "Not strictly sequence model.",
        "align": "Not strictly sequence model.",
        "altclip": "Not strictly sequence model.",
        "beit": "Not strictly sequence model.",
        "bit": "Not strictly sequence model.",
        "blip": "Not strictly sequence model.",
        "blip-2": "Not strictly sequence model.",
        "bridgetower": "Not strictly sequence model.",
        "canine": "Cannot do straightforward look up of embeddings.",
        "clap": "Not strictly sequence model.",
        "clip": "Not strictly sequence model.",
        "clipseg": "Not strictly sequence model.",
        "chinese_clip": "Not strictly sequence model.",
        "conditional_detr": "Not strictly sequence model.",
        "convbert": "HF error.",
        "convnext": "Not strictly sequence model..",
        "convnextv2": "Not strictly sequence model..",
        "cpmant": "AttributeError: 'NoneType' object has no attribute 'dtype'",
        "cvt": "Not strictly sequence model.",
        "data2vec-audio": "Not strictly sequence model.",
        "data2vec-vision": "Not strictly sequence model.",
        "decision_transformer": "Raises NotImplementedError",
        "decision_transformer_gpt2": "Raises ValueError: Unrecognized model identifier",
        "deformable_detr": "Not strictly sequence model.",
        "deit": "Not strictly sequence model.",
        "deta": "Not strictly sequence model.",
        "dinat": "Not strictly sequence model.",
        "detr": "Not strictly sequence model.",
        "donut-swin": "Not strictly sequence model.",
        "dpr": "Not strictly sequence model.",
        "dpt": "Not strictly sequence model.",
        "esm": "Needs fixing/special handling of init parameters.",
        "efficientformer": "Not strictly sequence model.",
        "efficientnet": "Not strictly sequence model.",
        "ernie_m": "AttributeError: 'bool' object has no attribute 'to'.",
        "fsmt": "Not strictly sequence model.",
        "focalnet": "Not strictly sequence model.",
        "funnel": "HF error.",
        "flava": "Not strictly sequence model.",
        "nat": "Requires the natten library",
        "glpn": "Not strictly sequence model.",
        "graphormer": "Not strictly sequence model.",
        "groupvit": "Not strictly sequence model.",
        "gpt_neo": "Configuration troublesome w.r.t. attn layers matching num_layers.",
        "gptsan-japanese": "AttributeError: 'MoECausalLMOutputWithPast' object "
        "has no attribute 'last_hidden_state'.",
        "hubert": "Cannot do straightforward look up of embeddings.",
        "informer": "Not strictly sequence model.",
        "jukebox": "Not strictly sequence model.",
        "layoutlmv2": "Not strictly sequence model.",
        "layoutlmv3": "Not strictly sequence model.",
        "levit": "Not strictly sequence model.",
        "lilt": "Matmul incompatible shapes with default params.",
        "lxmert": "Not strictly sequence model.",
        "maskformer": "Not strictly sequence model.",
        "mask2former": "Not strictly sequence model.",
        "maskformer-swin": "Not strictly sequence model.",
        "mctct": "get_input_embeddings() raises NotImplementedError.",
        "mgp-str": "Not strictly sequence model.",
        "mobilenet_v1": "Not strictly sequence model.",
        "mobilenet_v2": "Not strictly sequence model.",
        "mobilevit": "Not strictly sequence model.",
        "mt5": "Not implemented in EIR for feature extraction yet.",
        "oneformer": "Not strictly sequence model.",
        "owlvit": "Not strictly sequence model.",
        "nllb": "Throws ValueError: Unrecognized model identifier: nllb",
        "retribert": "Cannot do straightforward look up of embeddings.",
        "resnet": "Not strictly sequence model.",
        "regnet": "Not strictly sequence model.",
        "sam": "Not strictly sequence model.",
        "segformer": "Not strictly sequence model.",
        "sew": "Not strictly sequence model.",
        "sew-d": "Not strictly sequence model.",
        "speech_to_text": "Not strictly sequence model.",
        "speecht5": "Not strictly sequence model.",
        "swin": "Not strictly sequence model.",
        "swinv2": "Not strictly sequence model.",
        "swin2sr": "Not strictly sequence model.",
        "table-transformer": "Not strictly sequence model.",
        "tapas": "TapasModel requires the torch-scatter library.",
        "time_series_transformer": "Not strictly sequence model.",
        "timesformer": "Not strictly sequence model.",
        "trajectory_transformer": "Not strictly sequence model.",
        "tvlt": "Not strictly sequence model.",
        "perceiver": "Not strictly sequence model.",
        "poolformer": "Not strictly sequence model.",
        "qdqbert": "ImportError.",
        "unispeech": "Not strictly sequence model.",
        "unispeech-sat": "Not strictly sequence model.",
        "van": "Not strictly sequence model.",
        "videomae": "Not strictly sequence model.",
        "vilt": "Not strictly sequence model.",
        "vit": "Not strictly sequence model.",
        "vit_hybrid": "Not strictly sequence model.",
        "vit_mae": "Not strictly sequence model.",
        "vit_msn": "Not strictly sequence model.",
        "vision-text-dual-encoder": "Not strictly sequence model.",
        "xclip": "Not strictly sequence model.",
        "xmod": "ValueError: Input language unknown. Please call "
        "`XmodPreTrainedModel.set_default_language().",
        "wav2vec2": "Not strictly sequence model.",
        "wav2vec2-conformer": "Not strictly sequence model.",
        "wavlm": "NotImplementedError.",
        "whisper": "Not strictly sequence model.",
        "yolos": "Not strictly sequence model..",
    }

    return unsupported



from dataclasses import dataclass
from typing import Dict, Sequence

import pandas as pd
from sklearn.preprocessing import StandardScaler

from eir.data_load import label_setup
from eir.data_load.label_setup import (
    al_target_label_dict,
    al_label_transformers,
    TabularFileInfo,
    al_all_column_ops,
    transform_label_df,
)
from eir.experiment_io.experiment_io import load_transformers
from eir.predict_modules.predict_input_setup import prep_missing_con_dict
from eir.setup.config import Configs
from eir.target_setup.target_label_setup import (
    df_to_nested_dict,
    get_tabular_target_file_infos,
)


@dataclass
class PredictTargetLabels:
    label_dict: al_target_label_dict
    label_transformers: Dict[str, al_label_transformers]

    @property
    def all_labels(self):
        return self.label_dict


def get_target_labels_for_predict(
    output_folder: str,
    tabular_file_infos: Dict[str, TabularFileInfo],
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> PredictTargetLabels:
    df_labels_test = pd.DataFrame(index=ids)
    label_transformers = {}
    con_columns = []
    cat_columns = []

    for output_name, tabular_info in tabular_file_infos.items():
        all_columns = list(tabular_info.cat_columns) + list(tabular_info.con_columns)
        if not all_columns:
            raise ValueError(f"No columns specified in {tabular_file_infos}.")

        con_columns += tabular_info.con_columns
        cat_columns += tabular_info.cat_columns

        df_cur_labels = _load_labels_for_predict(
            tabular_info=tabular_info,
            ids_to_keep=ids,
            custom_label_ops=custom_column_label_parsing_ops,
        )
        df_cur_labels["Output Name"] = output_name

        df_labels_test = pd.concat((df_labels_test, df_cur_labels))

        cur_transformers = load_transformers(
            output_folder=output_folder,
            transformers_to_load={output_name: all_columns},
        )
        label_transformers[output_name] = cur_transformers[output_name]

    df_labels_test = df_labels_test.set_index("Output Name", append=True)
    df_labels_test = df_labels_test.dropna(how="all")

    labels_dict = parse_labels_for_predict(
        con_columns=con_columns,
        cat_columns=cat_columns,
        df_labels_test=df_labels_test,
        all_output_label_transformers=label_transformers,
    )

    labels = PredictTargetLabels(
        label_dict=labels_dict, label_transformers=label_transformers
    )

    return labels


def _load_labels_for_predict(
    tabular_info: TabularFileInfo,
    ids_to_keep: Sequence[str],
    custom_label_ops: al_all_column_ops = None,
) -> pd.DataFrame:
    parse_wrapper = label_setup.get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_info.parsing_chunk_size
    )
    df_labels_test = parse_wrapper(
        label_file_tabular_info=tabular_info,
        ids_to_keep=ids_to_keep,
        custom_label_ops=custom_label_ops,
    )

    return df_labels_test


def parse_labels_for_predict(
    con_columns: Sequence[str],
    cat_columns: Sequence[str],
    df_labels_test: pd.DataFrame,
    all_output_label_transformers: Dict[str, al_label_transformers],
) -> al_target_label_dict:
    con_transformers = _extract_target_con_transformers(
        all_output_label_transformers=all_output_label_transformers,
        con_columns=con_columns,
    )
    train_con_column_means = prep_missing_con_dict(con_transformers=con_transformers)

    df_labels_test = label_setup.handle_missing_label_values_in_df(
        df=df_labels_test,
        cat_label_columns=cat_columns,
        con_label_columns=con_columns,
        con_manual_values=train_con_column_means,
        name="test_df",
    )

    assert len(all_output_label_transformers) > 0
    for _, output_transformer_set in all_output_label_transformers.items():
        df_labels_test = transform_label_df(
            df_labels=df_labels_test, label_transformers=output_transformer_set
        )

    test_labels_dict = df_to_nested_dict(df=df_labels_test)

    return test_labels_dict


def _extract_target_con_transformers(
    all_output_label_transformers: Dict[str, al_label_transformers],
    con_columns: Sequence[str],
) -> Dict[str, StandardScaler]:
    con_transformers = {}

    for output_name, output_transformer_set in all_output_label_transformers.items():
        for target_column, transformer_object in output_transformer_set.items():
            if target_column not in con_columns:
                continue

            assert target_column not in con_transformers
            assert isinstance(transformer_object, StandardScaler)

            con_transformers[target_column] = transformer_object

    assert len(con_transformers) == len(con_columns)

    return con_transformers


def get_target_labels_for_testing(
    configs_overloaded_for_predict: Configs,
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> PredictTargetLabels:
    """
    NOTE:   This can be extended to more tabular data, including other files, if we
            update the parameters slightly.
    """

    target_infos = get_tabular_target_file_infos(
        output_configs=configs_overloaded_for_predict.output_configs
    )

    target_labels = get_target_labels_for_predict(
        output_folder=configs_overloaded_for_predict.global_config.output_folder,
        tabular_file_infos=target_infos,
        custom_column_label_parsing_ops=custom_column_label_parsing_ops,
        ids=ids,
    )

    return target_labels



from typing import DefaultDict, TYPE_CHECKING

if TYPE_CHECKING:
    from eir.data_load.datasets import Sample


def add_id_to_samples(
    samples: DefaultDict[str, "Sample"], sample_id: str
) -> DefaultDict[str, "Sample"]:
    """
    This kind of weird function is used because in some cases, we cannot expect the
    target labels to have added samples, because we could be predicting on completely
    unknown samples without any target label data.

    Hence, we might have sparse modular data available for the samples, e.g. only omics
    for some samples, but only tabular data for others. So we want to ensure that the
    data is filled in.
    """
    if not samples[sample_id].sample_id:
        samples[sample_id].sample_id = sample_id
    else:
        assert samples[sample_id].sample_id == sample_id

    return samples



import warnings
from functools import lru_cache
from pathlib import Path
from typing import DefaultDict, Union, Callable, Generator, Set, TYPE_CHECKING

warnings.filterwarnings("ignore", message=".*newer version of deeplake.*")

import deeplake
from tqdm import tqdm

from eir.data_load.data_source_modules.common_utils import add_id_to_samples

if TYPE_CHECKING:
    from eir.data_load.datasets import Sample


@lru_cache()
def is_deeplake_dataset(data_source: str) -> bool:
    if Path(data_source, "dataset_meta.json").exists():
        return True

    return False


@lru_cache()
def load_deeplake_dataset(data_source: str) -> deeplake.Dataset:
    dataset = deeplake.load(
        path=data_source,
        read_only=True,
        verbose=False,
    )
    if "ID" not in dataset.tensors.keys():
        raise ValueError(
            f"DeepLake dataset at {data_source} does not have an ID tensor. "
            f"Please add one to the dataset."
        )
    return dataset


def get_deeplake_input_source_iterable(
    deeplake_dataset: deeplake.Dataset, inner_key: str
) -> Generator[deeplake.Tensor, None, None]:
    def _is_empty(x: deeplake.Tensor) -> bool:
        return x.shape[0] == 0

    for deeplake_sample in deeplake_dataset[inner_key]:
        if _is_empty(x=deeplake_sample):
            continue

        yield deeplake_sample


def add_deeplake_data_to_samples(
    input_source: str,
    input_name: str,
    deeplake_input_inner_key: str,
    samples: DefaultDict[str, "Sample"],
    data_loading_hook: Callable,
    ids_to_keep: Union[None, Set[str]],
) -> DefaultDict[str, "Sample"]:
    """
    For normal files in a folder, this is holding the paths, which we can think of as
    unique pointer to the sample data, for this source.

    In the DeepLake case, this is an integer index pointing to the sample.
    """
    assert deeplake_input_inner_key is not None, "Deeplake input inner key is None"

    def _is_empty(x: deeplake.Tensor) -> bool:
        return x.shape[0] == 0

    deeplake_ds = load_deeplake_dataset(data_source=input_source)
    if deeplake_input_inner_key not in deeplake_ds.tensors.keys():
        raise ValueError(
            f"Input key {deeplake_input_inner_key} not found in deeplake dataset "
            f"{input_source}. Available inputs in deeplake dataset are "
            f"{deeplake_ds.tensors.keys()}."
        )

    if ids_to_keep is None:
        pbar = tqdm(total=len(deeplake_ds), desc=input_name)
    else:
        pbar = tqdm(total=len(ids_to_keep), desc=input_name)

    for deeplake_sample in deeplake_ds:
        cur_input_from_sample = deeplake_sample[deeplake_input_inner_key]
        if _is_empty(x=cur_input_from_sample):
            continue

        sample_id = deeplake_sample["ID"].numpy().item()

        if ids_to_keep is not None and sample_id not in ids_to_keep:
            continue

        sample_data_pointer = deeplake_sample.index.values[0].value
        sample_data = data_loading_hook(sample_data_pointer)

        samples = add_id_to_samples(samples=samples, sample_id=sample_id)
        samples[sample_id].inputs[input_name] = sample_data

        pbar.update(1)

    return samples



from math import isclose
from pathlib import Path
from typing import Tuple, Union, Dict, TYPE_CHECKING, overload

import numpy as np
from aislib.misc_utils import get_logger
from ignite.contrib.handlers import (
    ConcatScheduler,
    CosineAnnealingScheduler,
    create_lr_scheduler_with_warmup,
)
from ignite.engine import Engine, Events
from matplotlib import pyplot as plt
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.optimizer import Optimizer

from eir.setup.schemas import GlobalConfig
from eir.train_utils.evaluation import validation_handler
from eir.train_utils.metrics import (
    read_metrics_history_file,
    get_average_history_filepath,
)
from eir.train_utils.utils import get_run_folder, validate_handler_dependencies

if TYPE_CHECKING:
    from eir.train_utils.train_handlers import HandlerConfig
    from eir.train import Experiment

logger = get_logger(name=__name__, tqdm_compatible=True)


@validate_handler_dependencies([validation_handler])
def attach_lr_scheduler(
    engine: Engine,
    lr_scheduler: Union[ConcatScheduler, CosineAnnealingScheduler, ReduceLROnPlateau],
    experiment: "Experiment",
) -> None:
    """
    We use Events.ITERATION_COMPLETED for the plateau lr_scheduler to be in sync with
    the evaluation handler (which runs on completed iteration as well). To make sure
    we start at the lower bound, we manually set the LR below (in the case of warmup
    being used).

    We use iteration started for the cycle lr_scheduler, to make sure we start at the
    lower bound of the learning rate when using warmup for the first iteration.
    """
    gc = experiment.configs.global_config

    if gc.lr_schedule in ["cycle", "cosine"]:
        engine.add_event_handler(
            event_name=Events.ITERATION_STARTED, handler=lr_scheduler
        )

    elif gc.lr_schedule == "plateau":
        if gc.warmup_steps:
            logger.debug("Setting first iteration optimizer LR to %.0e.", gc.lr_lb)
            update_optimizer_lr(lr=gc.lr_lb, optimizer=experiment.optimizer)

        step_scheduler_params = _get_reduce_lr_on_plateau_step_params(
            global_config=gc, optimizer=experiment.optimizer
        )
        engine.add_event_handler(
            event_name=Events.ITERATION_COMPLETED,
            handler=_step_reduce_on_plateau_scheduler,
            reduce_on_plateau_scheduler=lr_scheduler,
            **step_scheduler_params
        )
    else:
        raise ValueError()


def _get_reduce_lr_on_plateau_step_params(
    global_config: GlobalConfig, optimizer: Optimizer
) -> Dict:
    run_folder = get_run_folder(output_folder=global_config.output_folder)
    validation_history_fpath = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix="validation_"
    )

    warmup_steps = _get_warmup_steps_from_cla(
        warmup_steps_arg=global_config.warmup_steps, optimizer=optimizer
    )

    params = {
        "validation_history_fpath": validation_history_fpath,
        "optimizer": optimizer,
        "sample_interval": global_config.sample_interval,
        "lr_upper_bound": global_config.lr,
        "lr_lower_bound": global_config.lr_lb,
        "warmup_steps": warmup_steps,
    }

    return params


def set_up_lr_scheduler(
    handler_config: "HandlerConfig",
) -> Union[ConcatScheduler, CosineAnnealingScheduler, ReduceLROnPlateau]:
    exp = handler_config.experiment
    gc = exp.configs.global_config

    lr_lower_bound = gc.lr_lb
    lr_upper_bound = gc.lr

    def _get_cycle_iter_size(warmup_steps_: int) -> int:
        """
        Why this weird max(2, ...)? This is because the `ignite`
        `CosineAnnealingScheduler` expects at least 2 steps, so if we have fewer steps
        than the warmup period, we have max(2, negative number) = 2 for
        compatibility.
        """
        steps = len(exp.train_loader)
        if gc.lr_schedule == "cosine":
            steps = max(2, steps * gc.n_epochs - warmup_steps_)

        return steps

    def _get_total_num_events(n_epochs: int, iter_per_epoch: int) -> int:
        return n_epochs * iter_per_epoch

    if gc.lr_schedule in ["cycle", "cosine"]:
        warmup_steps = _get_warmup_steps_from_cla(
            warmup_steps_arg=gc.warmup_steps, optimizer=exp.optimizer
        )

        cycle_iter_size = _get_cycle_iter_size(warmup_steps_=warmup_steps)

        lr_scheduler, lr_scheduler_args = _get_cosine_lr_scheduler(
            optimizer=exp.optimizer,
            lr_upper_bound=lr_upper_bound,
            lr_lower_bound=lr_lower_bound,
            cycle_iter_size=cycle_iter_size,
            schedule=gc.lr_schedule,
        )

        if warmup_steps:
            lr_scheduler, lr_scheduler_args = _attach_warmup_to_scheduler(
                lr_scheduler=lr_scheduler,
                lr_lower_bound=lr_lower_bound,
                lr_upper_bound=lr_upper_bound,
                duration=warmup_steps,
            )

        num_events = _get_total_num_events(
            n_epochs=gc.n_epochs, iter_per_epoch=len(exp.train_loader)
        )

        if gc.plot_lr_schedule:
            _plot_lr_schedule(
                lr_scheduler=lr_scheduler,
                num_events=num_events,
                output_folder=handler_config.run_folder,
                lr_scheduler_args=lr_scheduler_args,
            )

    elif gc.lr_schedule == "plateau":
        logger.debug("Plateau patience set to %d.", gc.lr_plateau_patience)

        """
        For compatibility with ignite EarlyStopping handler, we reduce the plateau
        patience steps passed into the ReduceLROnPlateau constructor. This is because
        EarlyStopping uses bad_steps >= patience, while ReduceLROnPlateau uses bad_steps
        > patience. In order to have a common interface when passing in any type of
        patience steps, we are going to reduce the passed in steps here.
        """
        patience_steps = gc.lr_plateau_patience - 1
        lr_scheduler = ReduceLROnPlateau(
            optimizer=exp.optimizer,
            mode="max",
            factor=gc.lr_plateau_factor,
            patience=patience_steps,
            min_lr=lr_lower_bound,
        )

    else:
        raise ValueError()

    return lr_scheduler


def _get_cosine_lr_scheduler(
    optimizer: Optimizer,
    lr_upper_bound: float,
    lr_lower_bound: float,
    cycle_iter_size: int,
    schedule: str,
) -> Tuple[CosineAnnealingScheduler, Dict]:
    """
    We return the arguments because the simulate_values are classmethods, which
    we need to pass the arguments too.
    """

    cosine_scheduler_kwargs = {
        "optimizer": optimizer,
        "param_name": "lr",
        "start_value": lr_upper_bound,
        "end_value": lr_lower_bound,
        "cycle_size": cycle_iter_size,
    }

    if schedule == "cycle":
        cosine_scheduler_kwargs["cycle_mult"] = 2
        cosine_scheduler_kwargs["start_value_mult"] = 1

    lr_scheduler = CosineAnnealingScheduler(**cosine_scheduler_kwargs)

    return lr_scheduler, cosine_scheduler_kwargs


@overload
def _get_warmup_steps_from_cla(warmup_steps_arg: None, optimizer: Optimizer) -> None:
    ...


@overload
def _get_warmup_steps_from_cla(warmup_steps_arg: str, optimizer: Optimizer) -> int:
    ...


def _get_warmup_steps_from_cla(warmup_steps_arg, optimizer):
    if warmup_steps_arg is None:
        return 0
    elif warmup_steps_arg == "auto":
        auto_steps = _calculate_auto_warmup_steps(optimizer=optimizer)
        logger.debug(
            "Using calculated %d steps for learning rate warmup due to 'auto' "
            "option for warmup.",
            auto_steps,
        )
        return auto_steps
    else:
        return int(warmup_steps_arg)


def _calculate_auto_warmup_steps(optimizer: Optimizer) -> int:
    def _calc_steps(b2):
        return round(2 / (1 - b2))

    first_param_group = optimizer.param_groups[0]
    if "betas" in first_param_group:
        b2_value = first_param_group["betas"][1]
        return _calc_steps(b2_value)

    return 2000


def _plot_lr_schedule(
    lr_scheduler: Union[ConcatScheduler, CosineAnnealingScheduler],
    num_events: int,
    lr_scheduler_args: Dict,
    output_folder: Path,
):
    simulated_vals = np.array(
        lr_scheduler.simulate_values(num_events=num_events, **lr_scheduler_args)
    )

    plt.plot(simulated_vals[:, 0], simulated_vals[:, 1])
    plt.title("Learning Rate Schedule")
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")

    plt.savefig(output_folder / "lr_schedule.pdf")
    plt.close("all")


def _attach_warmup_to_scheduler(
    lr_scheduler: CosineAnnealingScheduler,
    lr_lower_bound: float,
    lr_upper_bound: float,
    duration: int,
) -> Tuple[ConcatScheduler, Dict]:
    """
    We have `patched_duration_for_cosine_end` because attaching a warmup to a cosine
    scheduler seems to 'offset' it's length by one step. This means that if we have
    100 warmup steps, and 125 cosine steps, iteration 225 will be in the next cycle
    of a cosine annealing schedule, meaning the LR upper bound.

    This can be seen by monitoring the iteration and current optimizer LR, which shows
    e.g. (LR=0.01 iteration=225) when printing the values.

    This happens because internally, the `create_lr_scheduler_with_warmup` `ignite`
    function reduces the duration of the warmup by one step. That is if we pass in 100
    warmup steps, steps 1-98 will be in the warmup phase (99 steps in total). This
    happens because of the following line in `ignite` 0.3:

    milestones_values[-1] = (warmup_duration - 2, warmup_end_value - d)

    So since the warmup steps are 1 less, it means that the cosine annealing will be
    one step more, meaning that the last step is a new cycle.

    This can be tested quite easily with the `output_simulated_values` argument,
    which shows the jump in LR in the last cycle if we pass in `durations` unpatched.
    """

    patched_duration_for_cosine_end = duration + 1
    scheduler_w_warmup = create_lr_scheduler_with_warmup(
        lr_scheduler=lr_scheduler,
        warmup_start_value=lr_lower_bound,
        warmup_end_value=lr_upper_bound,
        warmup_duration=patched_duration_for_cosine_end,
    )

    concat_scheduler_args = {
        "schedulers": scheduler_w_warmup.schedulers,
        "durations": [patched_duration_for_cosine_end],
    }

    return scheduler_w_warmup, concat_scheduler_args


def _step_reduce_on_plateau_scheduler(
    engine: Engine,
    optimizer: Optimizer,
    lr_upper_bound: float,
    lr_lower_bound: float,
    sample_interval: int,
    reduce_on_plateau_scheduler: ReduceLROnPlateau,
    validation_history_fpath: Path,
    warmup_steps: Union[None, int],
) -> None:
    """
    We do the warmup manually here because currently ignite does not support warmup
    with ReduceLROnPlateau through create_lr_scheduler_with_warmup because
    ReduceLROnPlateau does not inherit from _LRScheduler.
    """
    iteration = engine.state.iteration

    if warmup_steps is not None and iteration <= warmup_steps:
        cur_lr = calculate_lr_after_linear_step(
            lr_start=lr_lower_bound,
            lr_end=lr_upper_bound,
            warmup_steps=warmup_steps,
            iteration=iteration,
        )
        update_optimizer_lr(lr=cur_lr, optimizer=optimizer)

    else:
        prev_lr = get_optimizer_lr(optimizer=optimizer)
        cur_bad_steps = reduce_on_plateau_scheduler.num_bad_epochs

        if iteration % sample_interval == 0 and not isclose(prev_lr, lr_lower_bound):
            validation_df = read_metrics_history_file(
                file_path=validation_history_fpath
            )
            latest_val_performance = validation_df["perf-average"].iloc[-1]

            reduce_on_plateau_scheduler.step(metrics=latest_val_performance)

            # See comment in set_up_lr_scheduler for +1 explanation
            streamlined_patience = reduce_on_plateau_scheduler.patience + 1
            _log_plateu_bad_step(
                iteration=iteration,
                prev_bad_steps=cur_bad_steps,
                cur_bad_steps=reduce_on_plateau_scheduler.num_bad_epochs,
                patience=streamlined_patience,
            )

            new_lr = get_optimizer_lr(optimizer=optimizer)
            _log_reduce_on_plateu_step(
                reduce_on_plateau_scheduler=reduce_on_plateau_scheduler,
                iteration=iteration,
                prev_lr=prev_lr,
                cur_lr=new_lr,
                patience=streamlined_patience,
            )


def update_optimizer_lr(lr: float, optimizer: Optimizer) -> None:
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def calculate_lr_after_linear_step(
    lr_start: float,
    lr_end: float,
    warmup_steps: int,
    iteration: int,
) -> float:
    step_size = (lr_end - lr_start) / warmup_steps
    cur_lr = lr_start + (step_size * iteration)

    return cur_lr


def get_optimizer_lr(optimizer: Optimizer) -> float:
    return optimizer.param_groups[0]["lr"]


def _log_plateu_bad_step(
    iteration: int, prev_bad_steps: int, cur_bad_steps: int, patience: int
) -> None:
    if cur_bad_steps > prev_bad_steps:
        logger.debug(
            "Iteration %d: Reduce LR On Plateau: %d / %d",
            iteration,
            cur_bad_steps,
            patience,
        )


def _log_reduce_on_plateu_step(
    reduce_on_plateau_scheduler: ReduceLROnPlateau,
    patience: int,
    iteration: int,
    prev_lr: float,
    cur_lr: float,
) -> None:
    """
    NOTE: The ReduceLROnPlateau works differently from ignite's EarlyStopping in the
    sense that EarlyStopping will trigger when bad steps are >= patience, while
    ReduceLROnPlateau will trigger when bad steps are *>* patience.
    """
    sched = reduce_on_plateau_scheduler

    if not isclose(prev_lr, cur_lr) and cur_lr > sched.min_lrs[0]:
        logger.info(
            "Iteration %d: Reduce LR On Plateau %d / %d. "
            "Reduced learning rate from %.0e to %.0e.",
            iteration,
            patience,
            patience,
            prev_lr,
            cur_lr,
        )



from functools import partial
from typing import Dict, Union, Callable, Type, Literal, TYPE_CHECKING

import torch
from torch import nn

from eir.data_load import data_utils
from eir.setup import schemas
from eir.train_utils.metrics import calculate_prediction_losses

if TYPE_CHECKING:
    from eir.setup.output_setup import al_output_objects_as_dict

al_cat_loss_names = Literal["CrossEntropyLoss"]
al_con_loss_names = Literal[
    "MSELoss",
    "L1Loss",
    "SmoothL1Loss",
    "PoissonNLLLoss",
    "GaussianNLLLoss",
]

al_cat_losses = nn.CrossEntropyLoss
al_con_losses = (
    nn.MSELoss | nn.L1Loss | nn.SmoothL1Loss | nn.PoissonNLLLoss | nn.GaussianNLLLoss
)

al_criteria = al_con_losses | al_cat_losses

al_criteria_dict = Dict[str, Dict[str, al_criteria]]

al_losses_classes = [
    Type[nn.CrossEntropyLoss],
    Type[nn.MSELoss],
    Type[nn.L1Loss],
    Type[nn.SmoothL1Loss],
    Type[nn.PoissonNLLLoss],
    Type[nn.GaussianNLLLoss],
]


def get_criteria(outputs_as_dict: "al_output_objects_as_dict") -> al_criteria_dict:
    criteria_dict = {}
    target_columns_gen = data_utils.get_output_info_generator(
        outputs_as_dict=outputs_as_dict
    )

    for output_name, column_type, column_name in target_columns_gen:
        output_config = outputs_as_dict[output_name].output_config
        label_smoothing = _get_label_smoothing(
            output_config=output_config,
            column_type=column_type,
        )

        loss_name = _parse_loss_name(
            output_config=output_config, column_type=column_type
        )

        criterion = get_criterion(
            column_type_=column_type,
            loss_name=loss_name,
            cat_label_smoothing_=label_smoothing,
        )

        if output_name not in criteria_dict:
            criteria_dict[output_name] = {}
        criteria_dict[output_name][column_name] = criterion

    return criteria_dict


def build_loss_dict() -> dict[str, list[str]]:
    loss_dict = {
        "cat": [
            "CrossEntropyLoss",
        ],
        "con": [
            "MSELoss",
            "L1Loss",
            "SmoothL1Loss",
            "PoissonNLLLoss",
            "GaussianNLLLoss",
        ],
    }

    return loss_dict


def get_criterion(
    column_type_: str, loss_name: str, cat_label_smoothing_: float = 0.0
) -> Union[nn.CrossEntropyLoss, Callable]:
    loss_dict = build_loss_dict()

    match column_type_, loss_name:
        case "con", _:
            assert loss_name in loss_dict["con"]

            loss_module = getattr(nn, loss_name)
            return partial(_calc_con_loss, loss_func=loss_module())

        case "cat", "CrossEntropyLoss":
            return nn.CrossEntropyLoss(label_smoothing=cat_label_smoothing_)

        case "cat", _:
            assert loss_name in loss_dict["cat"]
            loss_module = getattr(nn, loss_name)
            return loss_module()


def _parse_loss_name(
    output_config: schemas.OutputConfig, column_type: str
) -> al_cat_loss_names | al_con_loss_names:
    match column_type:
        case "cat":
            return output_config.output_type_info.cat_loss_name
        case "con":
            return output_config.output_type_info.con_loss_name
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _get_label_smoothing(
    output_config: schemas.OutputConfig,
    column_type: str,
) -> float:
    match column_type:
        case "con":
            return 0.0
        case "cat":
            return output_config.output_type_info.cat_label_smoothing
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _calc_con_loss(input: torch.Tensor, target: torch.Tensor, loss_func: al_con_losses):
    return loss_func(input=input.squeeze(), target=target.squeeze())


def get_loss_callable(criteria: al_criteria_dict):
    single_task_loss_func = partial(calculate_prediction_losses, criteria=criteria)
    return single_task_loss_func



from dataclasses import dataclass
from pathlib import Path
from typing import Sequence, Union, Dict, TYPE_CHECKING

from eir.data_load.label_setup import (
    Labels,
    set_up_train_and_valid_tabular_data,
    TabularFileInfo,
)
from eir.models.tabular.tabular import get_unique_values_from_transformers
from eir.setup import schemas

if TYPE_CHECKING:
    from eir.train_utils.step_logic import Hooks


@dataclass
class TabularInputInfo:
    labels: Labels
    input_config: schemas.InputConfig
    total_num_features: int


def set_up_tabular_input_for_training(
    input_config: schemas.InputConfig,
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    hooks: Union["Hooks", None],
) -> TabularInputInfo:
    tabular_file_info = get_tabular_input_file_info(
        input_source=input_config.input_info.input_source,
        tabular_data_type_config=input_config.input_type_info,
    )

    custom_ops = hooks.custom_column_label_parsing_ops if hooks else None
    tabular_labels = set_up_train_and_valid_tabular_data(
        tabular_file_info=tabular_file_info,
        custom_label_ops=custom_ops,
        train_ids=train_ids,
        valid_ids=valid_ids,
        include_missing=True,
    )

    total_num_features = get_tabular_num_features(
        label_transformers=tabular_labels.label_transformers,
        cat_columns=input_config.input_type_info.input_cat_columns,
        con_columns=input_config.input_type_info.input_con_columns,
    )
    tabular_input_info = TabularInputInfo(
        labels=tabular_labels,
        input_config=input_config,
        total_num_features=total_num_features,
    )

    return tabular_input_info


def get_tabular_input_file_info(
    input_source: str,
    tabular_data_type_config: schemas.TabularInputDataConfig,
) -> TabularFileInfo:
    table_info = TabularFileInfo(
        file_path=Path(input_source),
        con_columns=tabular_data_type_config.input_con_columns,
        cat_columns=tabular_data_type_config.input_cat_columns,
        parsing_chunk_size=tabular_data_type_config.label_parsing_chunk_size,
    )

    return table_info


def get_tabular_num_features(
    label_transformers: Dict, cat_columns: Sequence[str], con_columns: Sequence[str]
) -> int:
    unique_cat_values = get_unique_values_from_transformers(
        transformers=label_transformers,
        keys_to_use=cat_columns,
    )
    cat_num_features = sum(
        len(unique_values) for unique_values in unique_cat_values.values()
    )
    total_num_features = cat_num_features + len(con_columns)

    return total_num_features



import reprlib
import typing
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Type, Dict, Union, Sequence, Tuple, Any

import torch
from aislib.misc_utils import get_logger
from torch import nn

from eir.models import al_meta_model

logger = get_logger(name=__name__)


def load_model(
    model_path: Path,
    model_class: Type[nn.Module],
    model_init_kwargs: Dict,
    device: str,
    test_mode: bool,
    state_dict_keys_to_keep: Union[None, Sequence[str]] = None,
    state_dict_key_rename: Union[None, Sequence[Tuple[str, str]]] = None,
    strict_shapes: bool = True,
) -> Union[al_meta_model, nn.Module]:
    model = model_class(**model_init_kwargs)

    model = _load_model_weights(
        model=model,
        model_state_dict_path=model_path,
        device=device,
        state_dict_keys_to_keep=state_dict_keys_to_keep,
        state_dict_key_rename=state_dict_key_rename,
        strict_shapes=strict_shapes,
    )

    if test_mode:
        model.eval()

    return model


def _load_model_weights(
    model: nn.Module,
    model_state_dict_path: Path,
    device: str,
    state_dict_keys_to_keep: Union[None, Sequence[str]] = None,
    state_dict_key_rename: Union[None, Sequence[Tuple[str, str]]] = None,
    strict_shapes: bool = True,
) -> nn.Module:
    loaded_weights_state_dict = torch.load(model_state_dict_path, map_location=device)

    if state_dict_keys_to_keep:
        no_keys_before = len(loaded_weights_state_dict)
        loaded_weights_state_dict = _filter_state_dict_keys(
            state_dict=loaded_weights_state_dict, keys_to_keep=state_dict_keys_to_keep
        )
        logger.info(
            "Extracting %d/%d modules for feature extractors: '%s' from %s.",
            len(loaded_weights_state_dict),
            no_keys_before,
            state_dict_keys_to_keep,
            model_state_dict_path,
        )

    if state_dict_key_rename:
        for replace_tuple in state_dict_key_rename:
            logger.debug(
                "Renaming '%s' in pretrained model to '%s' in current model.",
                replace_tuple[0],
                replace_tuple[1],
            )
            loaded_weights_state_dict = _replace_dict_key_names(
                dict_=loaded_weights_state_dict, replace_pattern=replace_tuple
            )

    if not strict_shapes:
        model_state_dict = model.state_dict()
        loaded_weights_state_dict = _filter_incompatible_parameter_shapes_for_loading(
            source_state_dict=model_state_dict,
            destination_state_dict=loaded_weights_state_dict,
        )

    incompatible_keys = model.load_state_dict(
        state_dict=loaded_weights_state_dict, strict=False
    )

    no_missing = len(incompatible_keys.missing_keys)
    no_unexpected = len(incompatible_keys.unexpected_keys)
    no_incompatible_keys = no_missing + no_unexpected
    if no_incompatible_keys > 0:
        repr_object = reprlib.Repr()
        repr_object.maxother = 256
        repr_object.maxstring = 256
        logger.info(
            "Encountered incompatible modules when loading model from '%s'.\n"
            "Missing keys: \n%s\n"
            "Unexpected keys: \n%s\n"
            "This is expected if you are loading select modules from a saved model, "
            "which means you can ignore this message. If you are loading a pre-trained"
            " model as-is, then this is most likely an error and something unexpected"
            "has changed between the pre-training and setting up the current model"
            "from the pre-trained one.",
            model_state_dict_path,
            repr_object.repr(incompatible_keys.missing_keys),
            repr_object.repr(incompatible_keys.unexpected_keys),
        )

    model = model.to(device=device)

    return model


def _replace_dict_key_names(
    dict_: Dict[str, Any], replace_pattern: Tuple[str, str]
) -> OrderedDict:
    renamed_dict = OrderedDict()

    for key, value in dict_.items():
        new_key = key.replace(*replace_pattern)
        renamed_dict[new_key] = value

    return renamed_dict


def _filter_state_dict_keys(
    state_dict: typing.OrderedDict[str, torch.nn.Parameter], keys_to_keep: Sequence[str]
) -> typing.OrderedDict[str, torch.nn.Parameter]:
    filtered_state_dict = OrderedDict()

    for module_name, module_parameter in state_dict.items():
        if any(key in module_name for key in keys_to_keep):
            filtered_state_dict[module_name] = module_parameter

    return filtered_state_dict


def _filter_incompatible_parameter_shapes_for_loading(
    source_state_dict: Dict[str, Any], destination_state_dict: Dict[str, Any]
) -> Dict[str, Any]:
    destination_state_dict = deepcopy(destination_state_dict)

    for k in destination_state_dict:
        if k in source_state_dict:
            if destination_state_dict[k].shape != source_state_dict[k].shape:
                logger.info(
                    f"Skipping loading of parameter: {k} "
                    f"due to incompatible shapes. "
                    f"Source shape: {source_state_dict[k].shape}. "
                    f"Destination shape: {destination_state_dict[k].shape}."
                )
                destination_state_dict[k] = source_state_dict[k]

    return destination_state_dict



from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib
import numpy as np
import pandas as pd

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib import gridspec

from aislib.misc_utils import get_logger, ensure_path_exists

if TYPE_CHECKING:
    from eir.interpretation.interpret_omics import (
        al_top_gradients_dict,
        al_scaled_grads_dict,
    )

logger = get_logger(name=__name__, tqdm_compatible=True)


def generate_snp_gradient_matrix(
    snp_df: pd.DataFrame, snp_indexes: np.ndarray
) -> np.ndarray:
    """
    TODO:
        Make this robust for .bim files later. Here we are basically mapping between
        the one hot coordinates and whether we are pulling out a reference or
        alternative (minor, A1) allele from the .bim / .snp file.
        Note that plink uses: <Variant ID>_<counted allele> in the .raw file, where
        'By default, A1 alleles are counted; this can be customized with
        --recode-allele.'

    TODO:
        Check the 'forms' of the recoded file and what each number there corresponds
        to w.r.t REF / HET / ALT. The columns positions in the .bim file are well
        defined, so we just have to make sure we are mapping correctly between the
        one-hot and genotype files (i.e. what exactly does [1,0,0,0] map to).
        See: https://www.cog-genomics.org/plink/1.9/formats#bim
        See: https://www.cog-genomics.org/plink/2.0/formats#bim
        See: http://zzz.bwh.harvard.edu/plink/dataman.shtml#recode
        'By default, the minor allele is assigned to be A1.'

    NOTE:
        Plink counts alternative (minor) alleles, while eigenstrat counts reference
        alleles. For now we are going to write this with additive format in mind (i.e
        0 = 0 copies of alternative (i.e. 2 copies of reference), 2 = 2 copies of
        alternative).

    NOTE:
        What we flag as "alternative" and "reference" can in theory be somewhat
        arbitrary, as long as we are grabbing the correct nucleotides.
    """
    snp_matrix = np.zeros((4, len(snp_indexes)), dtype=object)

    for matrix_col, snp_idx in enumerate(snp_indexes):
        cur_row = snp_df.iloc[snp_idx]
        snp_matrix[0, matrix_col] = cur_row.REF * 2  # 0 alt alleles counted
        snp_matrix[1, matrix_col] = cur_row.REF + cur_row.ALT  # 1 alt alleles counted
        snp_matrix[2, matrix_col] = cur_row.ALT * 2  # 2 alt alleles counted
        snp_matrix[3, matrix_col] = ""  # missing
    return snp_matrix


def plot_top_gradients(
    gathered_scaled_grads: "al_scaled_grads_dict",
    top_gradients_dict: "al_top_gradients_dict",
    df_snps: pd.DataFrame,
    output_folder: Path,
    fname: str = "top_snps.pdf",
    custom_ylabel: str = None,
):
    n_cls = len(top_gradients_dict.keys())
    classes = sorted(list(top_gradients_dict.keys()))

    fig = plt.figure(figsize=(n_cls * 4, n_cls * 2 + 1))
    gs = gridspec.GridSpec(n_cls, n_cls, wspace=0.2, hspace=0.2)

    snp_names = df_snps["VAR_ID"].array
    for grad_idx, col_name in enumerate(classes):
        cls_top_idxs = top_gradients_dict[col_name]["top_n_idxs"]
        cur_snp_matrix = generate_snp_gradient_matrix(df_snps, cls_top_idxs)

        for cls_idx, row_name in enumerate(classes):
            cur_grads = gathered_scaled_grads[col_name][row_name]

            cur_ax = plt.subplot(gs[cls_idx, grad_idx])
            cur_ax.imshow(cur_grads, vmin=0, vmax=1)

            ylabel = row_name if not custom_ylabel else custom_ylabel
            cur_ax.set_ylabel(ylabel)

            if cls_idx == n_cls - 1:
                top_snp_names = snp_names[cls_top_idxs]
                cur_ax.set_xticks(np.arange(len(top_snp_names)))
                cur_ax.set_xticklabels(top_snp_names)
                plt.setp(cur_ax.get_xticklabels(), rotation=90, ha="center")
            else:
                cur_ax.set_xticklabels([])
                cur_ax.set_xticks([])

            cur_ax.set_yticks(np.arange(4))
            cur_ax.set_yticklabels(["REF", "HET", "ALT", "MIS"])

            for snp_form in range(cur_snp_matrix.shape[0]):
                for snp_id in range(cur_snp_matrix.shape[1]):
                    cur_ax.text(
                        snp_id,
                        snp_form,
                        cur_snp_matrix[snp_form, snp_id],
                        ha="center",
                        va="center",
                        color="w",
                    )

    axs = fig.get_axes()
    for ax, col_title in zip(axs[::n_cls], classes):
        ax.set_title(f"Top {col_title} SNPs")

    for ax in axs:
        ax.label_outer()

    # plt.tight_layout gives a warning here, seems to be gs specific behavior
    gs.tight_layout(fig)
    plt.savefig(output_folder / fname, bbox_inches="tight")
    plt.close("all")


def plot_snp_manhattan_plots(
    df_snp_grads: pd.DataFrame, outfolder: Path, title_extra: str = ""
):
    df_snp_grads_copy = df_snp_grads.copy()

    attributions_columns = [
        i for i in df_snp_grads_copy.columns if i.endswith("_attributions")
    ]

    df_snp_grads_copy = df_snp_grads_copy.sort_values(by=["CHR_CODE", "BP_COORD"])

    for col in attributions_columns:
        label_name = col.split("_attributions")[0]

        ax, fig = _get_manhattan_axis_and_figure(
            df=df_snp_grads_copy,
            chr_column_name="CHR_CODE",
            attribution_column_name=col,
        )

        y_ticks = ax.get_yticks()
        y_axis_tick_spacing = y_ticks[1] - y_ticks[0]
        y_max = df_snp_grads_copy[col].max() + y_axis_tick_spacing
        ax.set_ylim(ymin=0.0, ymax=y_max)

        ax.set_xlabel("Chromosome")
        ax.set_ylabel("Attribution")

        ax.set_title(f"{label_name}{title_extra}")
        plt.tight_layout()

        out_path = outfolder / f"manhattan/{label_name}_manhattan.png"
        ensure_path_exists(path=out_path)
        plt.savefig(out_path, bbox_inches="tight")
        plt.close("all")


def _get_manhattan_axis_and_figure(
    df: pd.DataFrame,
    chr_column_name: str,
    attribution_column_name: str,
    color=None,
    figure_size=(12, 6),
    ar=90,
    gwas_sign_line=False,
    gwasp=5e-08,
    dotsize=8,
    valpha=1,
    axxlabel=None,
    axylabel=None,
    axlabelfontsize=9,
    axlabelfontname="Arial",
):
    """Adapted from https://github.com/reneshbedre/bioinfokit#manhatten-plot."""

    _x, _y = "Chromosome", r"Attribution"
    colors = (
        "#a7414a",
        "#282726",
        "#6a8a82",
        "#a37c27",
        "#563838",
        "#0584f2",
        "#f28a30",
        "#f05837",
        "#6465a5",
        "#00743f",
        "#be9063",
        "#de8cf0",
        "#888c46",
        "#c0334d",
        "#270101",
        "#8d2f23",
        "#ee6c81",
        "#65734b",
        "#14325c",
        "#704307",
        "#b5b3be",
        "#f67280",
        "#ffd082",
        "#ffd800",
        "#ad62aa",
        "#21bf73",
        "#a0855b",
        "#5edfff",
        "#08ffc8",
        "#ca3e47",
        "#c9753d",
        "#6c5ce7",
        "#a997df",
        "#513b56",
        "#590925",
        "#007fff",
        "#bf1363",
        "#f39237",
        "#0a3200",
        "#8c271e",
    )

    df["tpval"] = df[attribution_column_name]

    df["ind"] = range(len(df))

    if color is not None and len(color) == 2:
        color_1 = int(df[chr_column_name].nunique() / 2) * [color[0]]
        color_2 = int(df[chr_column_name].nunique() / 2) * [color[1]]
        if df[chr_column_name].nunique() % 2 == 0:
            color_list = list(reduce(lambda x, y: x + y, zip(color_1, color_2)))
        elif df[chr_column_name].nunique() % 2 == 1:
            color_list = list(reduce(lambda x, y: x + y, zip(color_1, color_2)))
            color_list.append(color[0])
    elif color is not None and len(color) == df[chr_column_name].nunique():
        color_list = color
    elif color is None:
        # select colors randomly from the list based in number of chr
        color_list = colors[: df[chr_column_name].nunique()]
    else:
        raise ValueError("Error in color argument.")

    xlabels = []
    xticks = []
    fig, ax = plt.subplots(figsize=figure_size)
    i = 0
    for label, df1 in df.groupby(chr_column_name):
        df1.plot(
            kind="scatter",
            x="ind",
            y="tpval",
            color=color_list[i],
            s=dotsize,
            alpha=valpha,
            ax=ax,
        )
        df1_max_ind = df1["ind"].iloc[-1]
        df1_min_ind = df1["ind"].iloc[0]
        xlabels.append(label)
        xticks.append((df1_max_ind - (df1_max_ind - df1_min_ind) / 2))
        i += 1

    # add GWAS significant line
    if gwas_sign_line is True:
        ax.axhline(y=-np.log10(gwasp), linestyle="--", color="#7d7d7d", linewidth=1)

    ax.margins(x=0)
    ax.margins(y=0)
    ax.set_xticks(xticks)
    ax.set_xticklabels(xlabels, rotation=ar)

    if axxlabel:
        _x = axxlabel
    if axylabel:
        _y = axylabel
    ax.set_xlabel(_x, fontsize=axlabelfontsize, fontname=axlabelfontname)
    ax.set_ylabel(_y, fontsize=axlabelfontsize, fontname=axlabelfontname)

    return ax, fig



from functools import partial
from typing import Type, Union, Dict

import torch
from torch import nn

from eir.models.fusion import fusion_mgmoe, fusion_default, fusion_identity
from eir.models.layers import ResidualMLPConfig


def pass_through_fuse(features: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    return features


def get_fusion_module(
    model_type: str,
    model_config: Union[
        ResidualMLPConfig, fusion_identity.IdentityConfig, fusion_mgmoe.MGMoEModelConfig
    ],
    modules_to_fuse: nn.ModuleDict,
    out_feature_per_feature_extractor: Dict[str, int],
) -> nn.Module:
    fusion_in_dim = _get_fusion_input_dimension(modules_to_fuse=modules_to_fuse)
    fusion_class = get_fusion_class(fusion_model_type=model_type)
    fusion_module = fusion_class(
        model_config=model_config,
        fusion_in_dim=fusion_in_dim,
        out_feature_per_feature_extractor=out_feature_per_feature_extractor,
    )

    return fusion_module


def _get_fusion_input_dimension(modules_to_fuse: nn.ModuleDict) -> int:
    fusion_in_dim = sum(i.num_out_features for i in modules_to_fuse.values())
    return fusion_in_dim


def get_fusion_class(
    fusion_model_type: str,
) -> Type[nn.Module]:
    if fusion_model_type == "mgmoe":
        return fusion_mgmoe.MGMoEModel
    elif fusion_model_type == "default":
        return fusion_default.FusionModule
    elif fusion_model_type == "identity":
        return fusion_identity.IdentityFusionModel
    elif fusion_model_type == "pass-through":
        return partial(
            fusion_identity.IdentityFusionModel, fusion_callable=pass_through_fuse
        )
    raise ValueError(f"Unrecognized fusion model type: {fusion_model_type}.")



from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union, Dict, Literal
from typing import Set, overload, TYPE_CHECKING, Sequence, Iterable, Any

import joblib
import torch
from aislib.misc_utils import ensure_path_exists
from sklearn.preprocessing import LabelEncoder, StandardScaler
from torch import nn

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_dict

# Aliases
al_tabular_input = Dict[str, Union[List[str], torch.Tensor]]
al_unique_embed_vals = Dict[str, Set[str]]
al_emb_lookup_dict = Dict[str, Dict[str, Dict[int, int]]]


@dataclass
class TabularModelConfig:

    """
    :param model_type:
         Which type of image model to use.

    :param model_init_config:
          Configuration / arguments used to initialise model.
    """

    model_init_config: "SimpleTabularModelConfig"
    model_type: Literal["tabular"] = "tabular"


@dataclass
class SimpleTabularModelConfig:
    """
    :param l1:
        L1 regularization applied to the embeddings for categorical tabular inputs.

    :param fc_layer:
        Whether to add a single fully-connected layer to the model, alternative
        to looking up and passing the inputs through directly.
    """

    l1: float = 0.00

    fc_layer: bool = False


class SimpleTabularModel(nn.Module):
    def __init__(
        self,
        model_init_config: SimpleTabularModelConfig,
        cat_columns: Sequence[str],
        con_columns: Sequence[str],
        unique_label_values_per_column: Dict[str, Set[str]],
        device: str,
    ):
        """
        Note: It would be more natural maybe to do the lookup here
        (using self.embeddings_dict), but then we also have to do all tensor
        preparation (e.g. mixing) here. Perhaps for now better to do it outside in
        preparation hook. However, this way also keeps a common interface between all
        current models, where they are taking a tensor as input (otherwise, we would
        be taking a dict here).

        Note: Currently the only thing this model does is looking up the trainable
        embeddings. Other types of tabular models can be added later, with their own
        layers.
        """

        super().__init__()

        self.cat_columns = cat_columns
        self.con_columns = con_columns
        self.unique_label_values = unique_label_values_per_column
        self.device = device

        self.embeddings_dict = set_up_embedding_dict(
            unique_label_values=unique_label_values_per_column
        )

        emb_total_dim = 0
        con_total_dim = 0

        if self.embeddings_dict:
            emb_total_dim = attach_embeddings(self, self.embeddings_dict)
        if self.con_columns:
            con_total_dim = len(self.con_columns)

        self.input_dim = emb_total_dim + con_total_dim

        self.layer = nn.Identity()
        if model_init_config.fc_layer:
            self.layer = nn.Linear(
                in_features=self.input_dim, out_features=self.input_dim, bias=True
            )

    @property
    def num_out_features(self) -> int:
        return self.input_dim

    def script_submodules_for_tracing(self):
        for name, module in self.named_modules():
            if isinstance(module, nn.Embedding):
                scripted_module = torch.jit.script(obj=module)
                setattr(self, name, scripted_module)

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        if not self.cat_columns:
            return torch.empty(0)

        return torch.cat([torch.flatten(i) for i in self.parameters()])

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.layer(input)


def set_up_embedding_dict(
    unique_label_values: al_unique_embed_vals,
) -> al_emb_lookup_dict:
    emb_lookup_dict = {}

    for emb_col, emb_uniq_values in unique_label_values.items():
        sorted_unique_values = sorted(emb_uniq_values)
        lookup_table = {k: idx for idx, k in enumerate(sorted_unique_values)}
        emb_lookup_dict[emb_col] = {"lookup_table": lookup_table}

    return emb_lookup_dict


def attach_embeddings(model: nn.Module, embeddings_dict: al_emb_lookup_dict) -> int:
    total_emb_dimension = 0

    for emb_col in sorted(embeddings_dict.keys()):
        n_categories = len(embeddings_dict[emb_col]["lookup_table"])
        cur_embedding_dim = calc_embedding_dimension(n_categories=n_categories)

        embedding_module = nn.Embedding(
            num_embeddings=n_categories, embedding_dim=cur_embedding_dim
        )
        setattr(model, "embed_" + emb_col, embedding_module)

        total_emb_dimension += cur_embedding_dim

    return total_emb_dimension


def calc_embedding_dimension(n_categories: int) -> int:
    return min(600, round(1.6 * n_categories**0.56))


def get_tabular_inputs(
    input_cat_columns: Sequence[str],
    input_con_columns: Sequence[str],
    tabular_model: SimpleTabularModel,
    tabular_input: "al_tabular_input",
    device: str,
) -> Union[torch.Tensor, None]:
    """
    We want to have a wrapper function to gather all extra inputs needed by the model.
    """
    extra_embeddings = None
    if input_cat_columns:
        extra_embeddings = get_embeddings_from_labels(
            categorical_columns=input_cat_columns,
            labels=tabular_input,
            model=tabular_model,
        )

        if not input_con_columns:
            return extra_embeddings.to(device=device)

    extra_continuous = None
    if input_con_columns:
        extra_continuous = get_extra_continuous_inputs_from_labels(
            labels=tabular_input, continuous_columns=input_con_columns
        )

        if not input_cat_columns:
            return extra_continuous.to(device=device)

    if extra_continuous is not None and extra_embeddings is not None:
        concat_emb_and_con = torch.cat((extra_embeddings, extra_continuous), dim=1)
        return concat_emb_and_con.to(device=device)

    return None


def get_embeddings_from_labels(
    categorical_columns: Sequence[str],
    labels: Dict[str, Sequence[torch.Tensor]],
    model: SimpleTabularModel,
) -> torch.Tensor:
    """
    Note that the extra_embeddings is a list of tensors, where each tensor is a batch
    of embeddings for a given extra categorical column.

    Note also that we do not need categorical_columns here, we can grab them from the
    embeddings dict directly. But we use that so the order of embeddings in the
    concatenated tensor is clear.
    """

    if labels is None:
        raise ValueError("No extra labels found for when looking up embeddings.")

    extra_embeddings = []
    for col_key in categorical_columns:
        cur_embedding = lookup_embeddings(
            model=model,
            embedding_col=col_key,
            labels=labels[col_key],
        )
        extra_embeddings.append(cur_embedding)

    extra_embeddings = torch.cat(extra_embeddings, dim=1)

    return extra_embeddings


def lookup_embeddings(
    model: SimpleTabularModel,
    embedding_col: str,
    labels: Sequence[torch.Tensor],
) -> torch.Tensor:
    """
    This produces a batch of embeddings, with dimensions N x embed_dim.
    """

    cur_embedding_module: nn.Embedding = getattr(model, "embed_" + embedding_col)
    cur_embedding = cur_embedding_module(labels)

    return cur_embedding


def get_unique_embed_values(
    labels_dict: "al_label_dict", embedding_cols: Sequence[str]
) -> al_unique_embed_vals:
    unique_embeddings_dict = OrderedDict((i, set()) for i in sorted(embedding_cols))

    for sample_labels in labels_dict.values():
        for key, value in sample_labels.items():
            if key in embedding_cols:
                assert isinstance(value, int)
                unique_embeddings_dict[key].add(value)

    return unique_embeddings_dict


def get_extra_continuous_inputs_from_labels(
    labels: Dict[str, torch.Tensor], continuous_columns: Iterable[str]
) -> torch.Tensor:
    extra_continuous = []
    for col in continuous_columns:
        cur_con_labels = labels[col].unsqueeze(1).to(dtype=torch.float)
        extra_continuous.append(cur_con_labels)

    extra_continuous = torch.cat(extra_continuous, dim=1)

    return extra_continuous


@overload
def set_up_and_save_embeddings_dict(
    embedding_columns: None, labels_dict: "al_label_dict", run_folder: Path
) -> None:
    ...


@overload
def set_up_and_save_embeddings_dict(
    embedding_columns: List[str], labels_dict: "al_label_dict", run_folder: Path
) -> al_emb_lookup_dict:
    ...


def set_up_and_save_embeddings_dict(embedding_columns, labels_dict, run_folder):
    """
    We need to save it for test time models to be able to load.
    """
    if embedding_columns:
        embedding_dict = get_embedding_dict(
            labels_dict=labels_dict, embedding_cols=embedding_columns
        )
        outpath = run_folder / "extra_inputs" / "embeddings.save"
        ensure_path_exists(path=outpath)
        joblib.dump(value=embedding_dict, filename=outpath)
        return embedding_dict

    return None


def get_embedding_dict(
    labels_dict: "al_label_dict", embedding_cols: List[str]
) -> al_emb_lookup_dict:
    """
    Simple wrapper function to call other embedding functions to create embedding
    dictionary.
    """
    unique_embs = get_unique_embed_values(
        labels_dict=labels_dict, embedding_cols=embedding_cols
    )
    emb_lookup_dict = set_up_embedding_dict(unique_label_values=unique_embs)

    return emb_lookup_dict


def get_unique_values_from_transformers(
    transformers: Dict[str, Union[LabelEncoder, StandardScaler]],
    keys_to_use: Union[str, Iterable[str]],
) -> Dict[str, Any]:
    out = {}

    if not keys_to_use:
        return out

    if keys_to_use == "all":
        iterable = transformers.keys()
    else:
        iterable = keys_to_use

    for k in iterable:
        out[k] = set(transformers[k].classes_)

    return out



from pathlib import Path
from typing import Union, Optional

import numpy as np
import torch

from eir.data_load.data_preparation_modules.common import (
    _load_deeplake_sample,
    process_tensor_to_length,
)
from eir.data_load.data_source_modules import deeplake_ops
from eir.setup.input_setup_modules.setup_bytes import BytesInputInfo


def bytes_load_wrapper(
    data_pointer: Union[Path, int],
    input_source: str,
    dtype: str,
    deeplake_inner_key: Optional[str] = None,
) -> np.ndarray:
    if deeplake_ops.is_deeplake_dataset(data_source=input_source):
        assert deeplake_inner_key is not None
        bytes_data = _load_deeplake_sample(
            data_pointer=data_pointer,
            input_source=input_source,
            inner_key=deeplake_inner_key,
        ).astype(dtype=dtype)
    else:
        bytes_data = np.fromfile(file=data_pointer, dtype=dtype)

    return bytes_data


def prepare_bytes_data(
    bytes_input_object: "BytesInputInfo", bytes_data: np.ndarray, test_mode: bool
) -> torch.Tensor:
    """
    We use clone here to copy the original data, vs. using from_numpy
    which shares memory, causing us to modify the original data.
    """
    bio = bytes_input_object

    sampling_strat = bio.input_config.input_type_info.sampling_strategy_if_longer
    if test_mode:
        sampling_strat = "from_start"

    bytes_tensor = torch.LongTensor(bytes_data).detach().clone()

    padding_value = bio.vocab.get("<pad>", 0)
    cur_bytes_padded = process_tensor_to_length(
        tensor=bytes_tensor,
        max_length=bio.input_config.input_type_info.max_length,
        sampling_strategy_if_longer=sampling_strat,
        padding_value=padding_value,
    )

    return cur_bytes_padded



from collections import OrderedDict
from dataclasses import dataclass
from typing import Sequence, Union, Literal, Dict, TYPE_CHECKING

import torch
from aislib.pytorch_modules import Swish
from torch import nn

from eir.models.layers import MLPResidualBlock, ResidualMLPConfig
from eir.models.models_base import (
    create_multi_task_blocks_with_first_adaptor_block,
    merge_module_dicts,
    calculate_module_dict_outputs,
    construct_multi_branches,
    initialize_modules_from_spec,
    get_final_layer,
)

if TYPE_CHECKING:
    from eir.setup.output_setup import al_num_outputs_per_target


@dataclass
class TabularMLPResidualModelConfig(ResidualMLPConfig):

    """
    :param layers:
        Number of residual MLP residual blocks to use in the output module.

    :param fc_task_dim:
        Number of hidden nodes in each MLP residual block.

    :param rb_do:
        Dropout in each MLP residual block.

    :param fc_do:
        Dropout before final layer.

    :param stochastic_depth_p:
        Stochastic depth probability (probability of dropping input)
        for each residual block.

    :param final_layer_type:
        Which type of final layer to use to construct tabular output prediction.
    """

    final_layer_type: Union[Literal["linear"], Literal["mlp_residual"]] = "linear"


@dataclass
class TabularModelOutputConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use."""

    model_init_config: Union[TabularMLPResidualModelConfig]
    model_type: Literal["mlp_residual"] = "mlp_residual"


class ResidualTabularOutputModule(nn.Module):
    def __init__(
        self,
        model_config: TabularMLPResidualModelConfig,
        input_dimension: int,
        num_outputs_per_target: "al_num_outputs_per_target",
    ):
        super().__init__()

        self.model_config = model_config
        self.input_dimension = input_dimension
        self.num_outputs_per_target = num_outputs_per_target

        task_names = tuple(self.num_outputs_per_target.keys())

        task_resblocks_kwargs = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "stochastic_depth_p": self.model_config.stochastic_depth_p,
            "full_preactivation": False,
        }

        multi_task_branches = create_multi_task_blocks_with_first_adaptor_block(
            num_blocks=self.model_config.layers[0],
            branch_names=task_names,
            block_constructor=MLPResidualBlock,
            block_constructor_kwargs=task_resblocks_kwargs,
            first_layer_kwargs_overload={"in_features": self.input_dimension},
        )

        final_layers = get_default_tabular_output_final_layers(
            fc_task_dim=self.model_config.fc_task_dim,
            dropout_p=self.model_config.fc_do,
            num_outputs_per_target=self.num_outputs_per_target,
            task_names=task_names,
            final_layer_type=self.model_config.final_layer_type,
        )
        self.multi_task_branches = merge_module_dicts(
            (multi_task_branches, *final_layers)
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        output_modules_out = calculate_module_dict_outputs(
            input_=inputs, module_dict=self.multi_task_branches
        )

        return output_modules_out


def get_linear_final_act_spec(in_features: int, dropout_p: float):
    spec = OrderedDict(
        {
            "norm_final": (nn.LayerNorm, {"normalized_shape": in_features}),
            "act_final": (Swish, {}),
            "do_final": (nn.Dropout, {"p": dropout_p}),
        }
    )

    return spec


def get_default_tabular_output_final_layers(
    fc_task_dim: int,
    dropout_p: float,
    num_outputs_per_target: "al_num_outputs_per_target",
    task_names: Union[None, Sequence[str]],
    final_layer_type: Union[Literal["linear"], Literal["mlp_residual"]],
) -> Sequence[nn.Module]:
    final_layers = []
    if final_layer_type == "linear":
        if task_names is None:
            raise ValueError(
                "Task names are needed when using '%s' as final layer type.",
                final_layer_type,
            )

        final_act_spec = get_linear_final_act_spec(
            in_features=fc_task_dim, dropout_p=dropout_p
        )
        final_act = construct_multi_branches(
            branch_names=task_names,
            branch_factory=initialize_modules_from_spec,
            branch_factory_kwargs={"spec": final_act_spec},
        )
        final_layers.append(final_act)
        final_layer_specific_kwargs = {}

    elif final_layer_type == "mlp_residual":
        final_layer_specific_kwargs = {
            "dropout_p": dropout_p,
            "stochastic_depth_p": 0.0,
        }

    else:
        raise ValueError("Unknown final layer type: '%s'.", final_layer_type)

    final_layer = get_final_layer(
        in_features=fc_task_dim,
        num_outputs_per_target=num_outputs_per_target,
        layer_type=final_layer_type,
        layer_type_specific_kwargs=final_layer_specific_kwargs,
    )
    final_layers.append(final_layer)

    return tuple(final_layers)



from dataclasses import dataclass
from pathlib import Path

from eir.setup import schemas
from eir.setup.input_setup_modules.common import (
    DataDimensions,
    get_data_dimension_from_data_source,
)


@dataclass
class ArrayInputInfo:
    input_config: schemas.InputConfig
    data_dimensions: "DataDimensions"


def set_up_array_input(
    input_config: schemas.InputConfig, *args, **kwargs
) -> ArrayInputInfo:
    data_dimensions = get_data_dimension_from_data_source(
        data_source=Path(input_config.input_info.input_source),
        deeplake_inner_key=input_config.input_info.input_inner_key,
    )

    array_input_info = ArrayInputInfo(
        input_config=input_config,
        data_dimensions=data_dimensions,
    )

    return array_input_info



from dataclasses import dataclass, fields
from pathlib import Path
from typing import Sequence, Dict, Tuple

import numpy as np
from aislib.misc_utils import get_logger
from timm.models.registry import _model_pretrained_cfgs
from torchvision import transforms
from torchvision.datasets.folder import default_loader
from torchvision.transforms import Compose
from torchvision.transforms.functional import to_tensor

from eir.data_load.data_source_modules.deeplake_ops import (
    is_deeplake_dataset,
    load_deeplake_dataset,
    get_deeplake_input_source_iterable,
)
from eir.setup import schemas
from eir.setup.setup_utils import collect_stats

logger = get_logger(name=__name__)


@dataclass
class PretrainedImageModelInfo:
    url: str
    num_classes: int
    input_size: Sequence[int]
    pool_size: Sequence[int]
    mean: Sequence[float]
    std: Sequence[float]
    first_conv: str
    classifier: str


def get_timm_configs() -> Dict[str, PretrainedImageModelInfo]:
    default_configs = {}
    field_names = {i.name for i in fields(PretrainedImageModelInfo)}
    for name, dict_ in _model_pretrained_cfgs.items():
        common = {k: v for k, v in dict_.items() if k in field_names}
        default_configs[name] = PretrainedImageModelInfo(**common)

    return default_configs


@dataclass
class ImageInputInfo:
    input_config: schemas.InputConfig
    base_transforms: Compose
    all_transforms: Compose
    normalization_stats: "ImageNormalizationStats"
    num_channels: int


def set_up_image_input_for_training(
    input_config: schemas.InputConfig, *args, **kwargs
) -> ImageInputInfo:
    input_type_info = input_config.input_type_info

    num_channels = input_type_info.num_channels
    if not num_channels:
        num_channels = infer_num_channels(
            data_source=input_config.input_info.input_source,
            deeplake_inner_key=input_config.input_info.input_inner_key,
        )

    normalization_stats = get_image_normalization_values(input_config=input_config)

    base_transforms, all_transforms = get_image_transforms(
        target_size=input_config.input_type_info.size,
        normalization_stats=normalization_stats,
        auto_augment=input_type_info.auto_augment,
    )

    image_input_info = ImageInputInfo(
        input_config=input_config,
        base_transforms=base_transforms,
        all_transforms=all_transforms,
        normalization_stats=normalization_stats,
        num_channels=num_channels,
    )

    return image_input_info


def infer_num_channels(data_source: str, deeplake_inner_key: str) -> int:
    if is_deeplake_dataset(data_source=data_source):
        deeplake_ds = load_deeplake_dataset(data_source=data_source)
        deeplake_iter = get_deeplake_input_source_iterable(
            deeplake_dataset=deeplake_ds, inner_key=deeplake_inner_key
        )
        test_image_array = next(deeplake_iter).numpy()
        data_pointer = (
            f"[deeplake dataset {data_source}, input {deeplake_inner_key}, "
            f"image ID: {deeplake_ds['ID'][0].text()}]"
        )
    else:
        test_file = next(Path(data_source).iterdir())
        test_image = default_loader(path=str(test_file))
        test_image_array = np.array(test_image)
        data_pointer = test_file.name

    if test_image_array.ndim == 2:
        num_channels = 1
    else:
        num_channels = test_image_array.shape[-1]

    logger.info(
        "Inferring number of channels from source %s (using %s) as: %d",
        data_source,
        data_pointer,
        num_channels,
    )

    return num_channels


@dataclass
class ImageNormalizationStats:
    channel_means: Sequence[float]
    channel_stds: Sequence[float]


def get_image_normalization_values(
    input_config: schemas.InputConfig,
) -> ImageNormalizationStats:
    input_type_info = input_config.input_type_info

    pretrained_model_configs = get_timm_configs()

    means = input_type_info.mean_normalization_values
    stds = input_type_info.stds_normalization_values

    model_config = input_config.model_config

    if model_config.pretrained_model:
        cur_config = pretrained_model_configs[model_config.model_type]

        if not means:
            logger.info(
                "Using inferred image channel means (%s) from base on training "
                "statistics from pretrained '%s' model.",
                cur_config.mean,
                model_config.model_type,
            )
            means = cur_config.mean
        else:
            logger.warning(
                "Got manual values for channel means (%s) when using "
                "pretrained model '%s'. Usually one would use the means "
                "from the training data when '%s' was trained.",
                means,
                model_config.model_type,
                model_config.model_type,
            )
        if not stds:
            logger.info(
                "Using inferred image channel standard deviations (%s) from base on "
                "training statistics from pretrained '%s' model.",
                cur_config.std,
                model_config.model_type,
            )
            stds = cur_config.std
        else:
            logger.warning(
                "Got manual values for channel standard deviations (%s) "
                "when using pretrained model '%s'. Usually one would use "
                "the means from the training data when '%s' was trained.",
                stds,
                model_config.model_type,
                model_config.model_type,
            )
    else:
        if not means or not stds:
            input_source = input_config.input_info.input_source
            deeplake_inner_key = input_config.input_info.input_inner_key
            logger.info(
                "Not using a pretrained model and no mean and standard deviation "
                "statistics passed in. Gathering running image means and standard "
                "deviations from %s.",
                input_source,
            )

            if is_deeplake_dataset(data_source=input_source):
                deeplake_ds = load_deeplake_dataset(data_source=input_source)
                image_iter = get_deeplake_input_source_iterable(
                    deeplake_dataset=deeplake_ds, inner_key=deeplake_inner_key
                )
                tensor_iterator = (to_tensor(i.numpy()) for i in image_iter)
            else:
                file_iterator = Path(input_source).rglob("*")
                image_iterator = (default_loader(str(f)) for f in file_iterator)
                tensor_iterator = (to_tensor(i) for i in image_iterator)

            gathered_stats = collect_stats(tensor_iterable=tensor_iterator)
            means = gathered_stats.mean
            stds = gathered_stats.std
            logger.info(
                "Gathered the following means: %s and standard deviations: %s "
                "from %s.",
                means,
                stds,
                input_source,
            )

    stats = ImageNormalizationStats(channel_means=means, channel_stds=stds)

    return stats


def get_image_transforms(
    target_size: Sequence[int],
    normalization_stats: ImageNormalizationStats,
    auto_augment: bool,
) -> Tuple[Compose, Compose]:
    random_transforms = transforms.TrivialAugmentWide()
    target_resize = [int(i * 1.5) for i in target_size]

    base = [
        transforms.Resize(size=target_resize),
        transforms.CenterCrop(size=target_size),
        transforms.ToTensor(),
        transforms.Normalize(
            mean=normalization_stats.channel_means,
            std=normalization_stats.channel_stds,
        ),
    ]

    base_transforms = transforms.Compose(transforms=base)
    if auto_augment:
        logger.info("Image will be auto augmented with TrivialAugment during training.")
        all_transforms = transforms.Compose(transforms=[random_transforms] + base)
    else:
        all_transforms = base_transforms

    return base_transforms, all_transforms



from copy import deepcopy, copy
from pathlib import Path
from typing import (
    List,
    Tuple,
    Union,
    Dict,
    TYPE_CHECKING,
    Callable,
    Any,
    Sequence,
)

import matplotlib.pyplot as plt
import torch
from aislib.misc_utils import get_logger
from aislib.pytorch_modules import Swish
from ignite.engine import Engine
from ignite.handlers.lr_finder import FastaiLRFinder
from torch import nn
from torch.nn import Module
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

from eir.data_load.data_utils import get_output_info_generator
from eir.train_utils.utils import (
    call_hooks_stage_iterable,
)

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from eir.train import (  # noqa: F401
        Experiment,
        al_training_labels_batch,
        Batch,
    )
    from eir.train_utils.step_logic import (
        al_training_labels_target,
        al_input_batch,
        al_ids,
    )
    from eir.setup.output_setup import al_output_objects_as_dict

# Aliases
al_dloader_gathered_preds = Tuple[
    Dict[str, Dict[str, torch.Tensor]], "al_training_labels_target", List[str]
]
al_dloader_gathered_raw = Tuple[
    Dict[str, torch.Tensor], "al_training_labels_target", Sequence[str]
]
al_lr_find_results = Dict[str, List[Union[float, List[float]]]]

logger = get_logger(name=__name__, tqdm_compatible=True)


def predict_on_batch(
    model: Module, inputs: Dict[str, torch.Tensor]
) -> Dict[str, Dict[str, torch.Tensor]]:
    assert not model.training
    with torch.no_grad():
        val_outputs = model(inputs=inputs)

    return val_outputs


def parse_target_labels(
    output_objects: "al_output_objects_as_dict",
    device: str,
    labels: "al_training_labels_target",
) -> "al_training_labels_target":
    target_columns_gen = get_output_info_generator(outputs_as_dict=output_objects)

    labels_casted = {}

    for output_name, column_type, column_name in target_columns_gen:
        if output_name not in labels_casted:
            labels_casted[output_name] = {}

        cur_labels = labels[output_name][column_name]

        if column_type == "con":
            labels_casted[output_name][column_name] = cur_labels.to(
                dtype=torch.float, device=device
            )
        elif column_type == "cat":
            labels_casted[output_name][column_name] = cur_labels.to(
                dtype=torch.long, device=device
            )

    return labels_casted


def gather_pred_outputs_from_dloader(
    data_loader: DataLoader,
    batch_prep_hook: Sequence[Callable],
    batch_prep_hook_kwargs: Dict[str, Any],
    model: Module,
    with_labels: bool = True,
) -> al_dloader_gathered_preds:
    """
    Used to gather predictions from a dataloader, normally for evaluation – hence the
    assertion that we are in eval mode.

    Why the deepcopy when appending labels? See:
    https://github.com/pytorch/pytorch/issues/973#issuecomment-459398189

    TODO: Use hook model forward here.
    """
    all_output_batches = []
    all_label_batches = []
    ids_total = []

    assert not model.training
    for loader_batch in data_loader:
        state = call_hooks_stage_iterable(
            hook_iterable=batch_prep_hook,
            common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
            state=None,
        )
        batch = state["batch"]

        inputs = batch.inputs
        target_labels = batch.target_labels
        ids = batch.ids

        outputs = predict_on_batch(model=model, inputs=inputs)

        all_output_batches.append(outputs)

        ids_total += [i for i in deepcopy(ids)]

        if with_labels:
            all_label_batches.append(deepcopy(target_labels))

        del inputs
        del target_labels
        del ids

    if with_labels:
        all_label_batches = _stack_list_of_output_target_dicts(
            list_of_target_batch_dicts=all_label_batches
        )

    all_output_batches = _stack_list_of_output_target_dicts(
        list_of_target_batch_dicts=all_output_batches
    )

    return all_output_batches, all_label_batches, ids_total


def gather_data_loader_samples(
    data_loader: DataLoader,
    batch_prep_hook: Sequence[Callable],
    batch_prep_hook_kwargs: Dict[str, Any],
    n_samples: Union[int, None] = None,
) -> al_dloader_gathered_raw:
    all_input_batches = []
    all_label_batches = []
    ids_total = []

    for loader_batch in data_loader:
        state = call_hooks_stage_iterable(
            hook_iterable=batch_prep_hook,
            common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
            state=None,
        )
        batch = state["batch"]

        inputs: "al_input_batch" = batch.inputs
        target_labels: "al_training_labels_target" = batch.target_labels
        ids: "al_ids" = batch.ids

        all_input_batches.append(inputs)
        all_label_batches.append(target_labels)
        ids_total += [i for i in ids]

        if n_samples:
            if len(ids_total) >= n_samples:
                ids_total = ids_total[:n_samples]
                break

    all_label_batches = _stack_list_of_output_target_dicts(
        list_of_target_batch_dicts=all_label_batches
    )
    all_input_batches = _stack_list_of_batch_dicts(
        list_of_batch_dicts=all_input_batches
    )

    if n_samples:
        inputs_final, target_labels_final = {}, {}

        for input_name in all_input_batches.keys():
            input_subset = all_input_batches[input_name][:n_samples]
            inputs_final[input_name] = input_subset

        for output_name in all_label_batches.keys():
            cur_output = all_label_batches[output_name]
            for target_name in cur_output.keys():
                target_subset = cur_output[target_name][:n_samples]
                target_labels_final[target_name] = target_subset

    else:
        inputs_final, target_labels_final = all_input_batches, all_label_batches

    return inputs_final, target_labels_final, ids_total


def _stack_list_of_output_target_dicts(
    list_of_target_batch_dicts: List["al_training_labels_target"],
) -> "al_training_labels_target":
    """
    Spec:
        [batch_1, batch_2, batch_3]

        batch_1 =   {
                        'Output_Name_1': {'Target_Column_1': torch.Tensor(...)},
                        'Output_Name_2': {'Target_Column_2': torch.Tensor(...)},
                    }
                                                            with obs as rows in tensors
    """

    output_names = list_of_target_batch_dicts[0].keys()
    aggregated_batches = {output_name: {} for output_name in output_names}
    for output_name in output_names:
        cur_output_targets = list_of_target_batch_dicts[0][output_name]
        for target_name in cur_output_targets.keys():
            aggregated_batches[output_name][target_name] = []

    for batch in list_of_target_batch_dicts:
        assert set(batch.keys()) == output_names

        for output_name in batch.keys():
            cur_output_batch = batch[output_name]
            for target_name in cur_output_batch.keys():
                cur_column_batch = cur_output_batch[target_name]
                cur_batch_value = [i for i in cur_column_batch]

                aggregated_batches[output_name][target_name] += cur_batch_value

    stacked_outputs = {}
    for output_name, output_dict in aggregated_batches.items():
        cur_stacked_outputs = {
            key: _do_stack(list_of_elements)
            for key, list_of_elements in output_dict.items()
        }
        stacked_outputs[output_name] = cur_stacked_outputs

    return stacked_outputs


def _stack_list_of_batch_dicts(
    list_of_batch_dicts: List["al_input_batch"],
) -> "al_input_batch":
    """
    Spec:
        [batch_1, batch_2, batch_3]
        batch_1 =   {
                        'input_1': torch.Tensor(...), # with obs as rows
                        'input_2': torch.Tensor(...),
                    }
    """

    target_columns = list_of_batch_dicts[0].keys()
    aggregated_batches = {key: [] for key in target_columns}

    for batch in list_of_batch_dicts:
        assert set(batch.keys()) == target_columns

        for column in batch.keys():
            cur_column_batch = batch[column]
            aggregated_batches[column] += [i for i in cur_column_batch]

    stacked_inputs = {
        key: _do_stack(list_of_elements)
        for key, list_of_elements in aggregated_batches.items()
    }

    return stacked_inputs


def _do_stack(
    list_of_elements: List[Union[torch.Tensor, torch.LongTensor, str]]
) -> Union[torch.Tensor, List[str]]:
    # check that they're all the same type
    list_types = set(type(i) for i in list_of_elements)
    assert len(list_types) == 1

    are_tensors = isinstance(list_of_elements[0], (torch.Tensor, torch.LongTensor))
    if are_tensors:
        return torch.stack(list_of_elements)

    return list_of_elements


def add_wd_to_model_params(
    model: nn.Module, wd: float
) -> List[Dict[str, Union[str, float]]]:
    """
    We want to skip adding weight decay to learnable activation parameters so as
    not to bias them towards 0.

    TODO:   Split this function in two, one to get the parameters and one to add the
            WD to them. Possibly we have to do it in-place here, not copy as we have
            tensors.

    Note: Since we are adding the weight decay manually here, the optimizer does not
    touch the parameter group weight decay at initialization.
    """
    _check_named_modules(model=model)

    params = []
    for name, param in model.named_parameters():
        cur_dict = {"params": param}

        if "act_" in name:
            cur_dict["weight_decay"] = 0.0
        else:
            cur_dict["weight_decay"] = wd

        params.append(cur_dict)

    return params


def _check_named_modules(model: nn.Module):
    """
    We have this function as a safeguard to check that activations that have learnable
    parameters are named correctly (so that WD is not applied to them). Also, we want
    to make sure we don't have modules that are named 'incorrectly' and have the WD
    skipped when they should have it.
    """

    for name, module in model.named_modules():
        if name.startswith("act_"):
            assert isinstance(module, (Swish, nn.PReLU)), (name, module)

        if isinstance(module, (Swish, nn.PReLU)):
            assert "act_" in name, name


def run_lr_find(
    trainer_engine: Engine,
    train_dataloader: torch.utils.data.DataLoader,
    model: nn.Module,
    optimizer: Optimizer,
    output_folder: Path,
):
    lr_range_results, lr_suggestion = get_lr_range_results(
        trainer_engine=trainer_engine,
        train_dataloader=train_dataloader,
        model=model,
        optimizer=optimizer,
    )

    lr_range_results_parsed = _parse_out_lr_find_multiple_param_groups(
        lr_find_results=lr_range_results
    )
    lr_suggestion_parsed = _parse_out_lr_find_multiple_param_groups_suggestion(
        lr_finder_suggestion=lr_suggestion
    )

    plot_lr_find_results(
        lr_find_results=lr_range_results_parsed,
        lr_suggestion=lr_suggestion_parsed,
        output_folder=output_folder,
    )


def get_lr_range_results(
    trainer_engine: Engine,
    train_dataloader: torch.utils.data.DataLoader,
    model: nn.Module,
    optimizer: Optimizer,
    num_iter: int = 500,
) -> Tuple[al_lr_find_results, List[float]]:
    """
    We do a little hack with max_epochs and epoch_length because we don't pass that
    to the original trainer when running a normal training instance, which uses the
    default of max_epochs = 1. This is normally not a problem, as 1 epoch is
    generally more than for example 300 iterations, which should be enough for the LR
    test.

    However, in the cases where we have a small epoch length, the LR find will not
    run to completion as it only runs for one epoch. Hence we need to patch the
    max_epochs here for the test.
    """
    lr_finder = FastaiLRFinder()

    to_save = {"optimizer": optimizer, "model": model}
    with lr_finder.attach(
        trainer_engine,
        to_save=to_save,
        output_transform=lambda x: x["average"]["loss-average"],
        num_iter=num_iter,
    ) as trainer_with_lr_finder:
        logger.info("Running LR range test for max %d iterations.", num_iter)

        default_max_epochs = trainer_with_lr_finder.state.max_epochs
        trainer_with_lr_finder.state.max_epochs = 100
        trainer_with_lr_finder.state.epoch_length = len(train_dataloader)

        assert (
            trainer_with_lr_finder.state.max_epochs
            * trainer_with_lr_finder.state.epoch_length
        ) >= num_iter

        trainer_with_lr_finder.run(train_dataloader)
        trainer_with_lr_finder.state.max_epochs = default_max_epochs

        return deepcopy(lr_finder.get_results()), deepcopy(lr_finder.lr_suggestion())


def _parse_out_lr_find_multiple_param_groups(lr_find_results: al_lr_find_results):
    lr_find_results_copy = copy(lr_find_results)
    lr_steps = lr_find_results["lr"]
    logger.info(
        "LR Find: The model has %d parameter groups which can have different "
        "learning rates depending on implementation. For the LR find, "
        "we are using the LR defined in group 0, so make sure that is taken "
        "into consideration when analysing LR find results, the supplied LR "
        "for training might have to be scaled accordingly.",
        len(lr_steps[0]),
    )
    first_param_groups_lrs = [i[0] for i in lr_steps]

    lr_find_results_copy["lr"] = first_param_groups_lrs
    return lr_find_results_copy


def _parse_out_lr_find_multiple_param_groups_suggestion(
    lr_finder_suggestion: List[float],
):
    return lr_finder_suggestion[0]


def plot_lr_find_results(
    lr_find_results: al_lr_find_results,
    lr_suggestion: float,
    output_folder: Path,
):
    lr_values = copy(lr_find_results["lr"])
    loss_values = copy(lr_find_results["loss"])

    plt.plot(lr_values, loss_values)
    plt.xscale("log")
    plt.title("Learning Rate Search")
    plt.xlabel("Learning Rate")
    plt.ylabel("Loss")
    plt.axvline(x=lr_suggestion, color="red", linewidth=1)

    plt.savefig(str(output_folder / "lr_search.pdf"))


def trace_eir_model(
    meta_model: nn.Module, example_inputs: Dict[str, Any]
) -> torch.jit.TracedModule:
    """
    Optimally we would like to trace the whole meta_model in one go, but since torch
    currently does not like / support nested dict outputs when tracing a model,
    we'll opt for tracing the individual modules like below for now (assuming it's
    better than nothing).
    """

    meta_model.eval()

    for name, module in meta_model.named_modules():
        if hasattr(module, "script_submodules_for_tracing"):
            module.script_submodules_for_tracing()

    with torch.no_grad():
        traced_input_modules = nn.ModuleDict()
        feature_extractors_out = {}
        for module_name, module_input in example_inputs.items():
            module = meta_model.input_modules[module_name]
            traced_input_module = torch.jit.trace_module(
                mod=module,
                inputs={"forward": module_input},
                strict=False,
            )
            traced_input_modules[module_name] = traced_input_module
            feature_extractors_out[module_name] = module(module_input)

        fusion_module = meta_model.fusion_module
        traced_fusion_module = torch.jit.trace_module(
            mod=fusion_module,
            inputs={"forward": feature_extractors_out},
            strict=False,
        )
        fused_features = fusion_module(feature_extractors_out)

        traced_output_modules = nn.ModuleDict()
        for output_module_name, output_module in meta_model.output_modules.items():
            traced_output_module = torch.jit.trace_module(
                mod=output_module,
                inputs={"forward": fused_features},
                strict=False,
            )
            traced_output_modules[output_module_name] = traced_output_module

    traced_meta_model = meta_model.__class__(
        input_modules=traced_input_modules,
        fusion_module=traced_fusion_module,
        output_modules=traced_output_modules,
    )

    return traced_meta_model



from collections import OrderedDict
from dataclasses import dataclass
from typing import List, TYPE_CHECKING, Union

import torch
from aislib import pytorch_utils
from aislib.misc_utils import get_logger
from aislib.pytorch_modules import Swish
from torch import nn

from eir.models.layers import FirstCNNBlock, CNNResidualBlock, ConvAttentionBlock

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions

logger = get_logger(__name__)


@dataclass
class CNNModelConfig:
    """
    Note that when using the automatic network setup, channels will get increased as
    the input gets propagated through the network while the width and height get
    reduced due to stride.

    :param layers:
        Controls the number of layers in the model. If set to ``None``, the model will
        automatically set up the number of layers until a certain width and height
        (stride * 8 for width and height) are met. Future work includes adding a
        parameter to control the target width and height.

    :param fc_repr_dim:
        Output dimension of the last FC layer in the network which accepts the outputs
        from the convolutional layer.

    :param channel_exp_base:
        Which power of 2 to use in order to set the number of channels in the network.
        For example, setting ``channel_exp_base=3`` means that 2**3=8 channels will be
        used.

    :param first_channel_expansion:
        Factor to extend the first layer channels.

    :param kernel_width:
        Base kernel width of the convolutions.

    :param first_kernel_expansion_width:
        Factor to extend the first kernel's width.

    :param down_stride_width:
        Down stride of the convolutional layers along the width.

    :param first_stride_expansion_width:
        Factor to extend the first layer stride along the width.

    :param dilation_factor_width:
        Base dilation factor of the convolutions along the width in the network.

    :param kernel_height:
        Base kernel height of the convolutions.

    :param first_kernel_expansion_height:
        Factor to extend the first kernel's height.

    :param down_stride_height:
        Down stride of the convolutional layers along the height.

    :param first_stride_expansion_height:
        Factor to extend the first layer stride along the height.

    :param dilation_factor_height:
        Base dilation factor of the convolutions along the height in the network.

    :param cutoff:
        If the *resulting* dimension of width * height of adding a successive block
        is less than this value, will stop adding residual blocks to the
        model in the automated case (i.e., if the layers argument is not specified).

    :param rb_do:
        Dropout in the convolutional residual blocks.

    :param stochastic_depth_p:
        Probability of dropping input.

    :param attention_inclusion_cutoff:
        If the dimension of width * height is less than this value, attention will be
        included in the model across channels and width * height after that point.

    :param l1:
        L1 regularization to apply to the first layer.
    """

    layers: Union[None, List[int]] = None

    fc_repr_dim: int = 32

    channel_exp_base: int = 2
    first_channel_expansion: int = 1

    kernel_width: int = 12
    first_kernel_expansion_width: int = 1
    down_stride_width: int = 4
    first_stride_expansion_width: int = 1
    dilation_factor_width: int = 1

    kernel_height: int = 4
    first_kernel_expansion_height: int = 1
    down_stride_height: int = 1
    first_stride_expansion_height: int = 1
    dilation_factor_height: int = 1

    cutoff: int = 32

    rb_do: float = 0.00

    stochastic_depth_p: float = 0.00

    attention_inclusion_cutoff: int = 0

    l1: float = 0.00


def _validate_cnn_config(model_config: CNNModelConfig) -> None:
    mc = model_config

    if mc.down_stride_width > mc.kernel_width:
        raise ValueError(
            f"Down stride width {mc.down_stride_width}"
            f" is greater than kernel width {mc.kernel_width}."
            f"This is currently not supported."
        )

    if mc.down_stride_height > mc.kernel_height:
        raise ValueError(
            f"Down stride height {mc.down_stride_height} "
            f" is greater than kernel height {mc.kernel_height}."
            f"This is currently not supported."
        )

    first_stride_w = mc.down_stride_width * mc.first_stride_expansion_width
    first_kernel_w = mc.kernel_width * mc.first_kernel_expansion_width

    first_stride_h = mc.down_stride_height * mc.first_stride_expansion_height
    first_kernel_h = mc.kernel_height * mc.first_kernel_expansion_height

    if first_stride_w > first_kernel_w:
        raise ValueError(
            f"Effective down stride width {first_stride_w}"
            f" (down stride width {mc.down_stride_width}"
            f" times first stride expansion width {mc.first_stride_expansion_width})"
            f" is greater than effective kernel width {first_kernel_w}"
            f" (kernel width {mc.kernel_width}"
            f" times first kernel expansion width {mc.first_kernel_expansion_width})."
            f"This is currently not supported."
        )

    if first_stride_h > first_kernel_h:
        raise ValueError(
            f"Effective down stride height {first_stride_h} "
            f" (down stride height {mc.down_stride_height} "
            f"times first stride expansion height {mc.first_stride_expansion_height})"
            f" is greater than effective kernel height {first_kernel_h}"
            f" (kernel height {mc.kernel_height}"
            f" times first kernel expansion height {mc.first_kernel_expansion_height})."
            f"This is currently not supported."
        )

    positive_int_params = [
        "fc_repr_dim",
        "first_channel_expansion",
        "kernel_width",
        "first_kernel_expansion_width",
        "down_stride_width",
        "first_stride_expansion_width",
        "dilation_factor_width",
        "kernel_height",
        "first_kernel_expansion_height",
        "down_stride_height",
        "first_stride_expansion_height",
        "dilation_factor_height",
        "cutoff",
    ]
    for param in positive_int_params:
        if getattr(mc, param) <= 0:
            raise ValueError(f"{param} must be a positive integer.")

    float_params = ["rb_do", "stochastic_depth_p"]
    for param in float_params:
        value = getattr(mc, param)
        if not (0 <= value <= 1):
            raise ValueError(f"{param} must be in the range [0, 1].")

    if mc.l1 < 0:
        raise ValueError("l1 must be non-negative.")

    if mc.layers is not None:
        if not all(isinstance(layer, int) and layer > 0 for layer in mc.layers):
            raise ValueError(
                "layers must be a list of positive integers "
                "(or None for automatic setup)."
            )


class CNNModel(nn.Module):
    def __init__(
        self,
        model_config: CNNModelConfig,
        data_dimensions: "DataDimensions",
    ):
        super().__init__()

        _validate_cnn_config(model_config=model_config)

        self.model_config = model_config
        self.data_dimensions = data_dimensions

        self.pos_representation = GeneralPositionalEmbedding(
            embedding_dim=self.data_dimensions.height,
            max_length=self.data_dimensions.width,
            dropout=0.0,
        )

        self.conv = nn.Sequential(
            *_make_conv_layers(
                residual_blocks=self.residual_blocks,
                cnn_model_configuration=self.model_config,
                data_dimensions=self.data_dimensions,
            )
        )

        size_func = pytorch_utils.calc_size_after_conv_sequence
        size_after_conv_w, size_after_conv_h = size_func(
            input_width=self.data_dimensions.width,
            input_height=self.data_dimensions.height,
            conv_sequence=self.conv,
        )
        self.data_size_after_conv = size_after_conv_w * size_after_conv_h

        self.no_out_channels = self.conv[-1].out_channels

        self.fc = nn.Sequential(
            OrderedDict(
                {
                    "fc_1_norm_1": nn.LayerNorm(normalized_shape=self.fc_1_in_features),
                    "fc_1_act_1": Swish(),
                    "fc_1_linear_1": nn.Linear(
                        in_features=self.fc_1_in_features,
                        out_features=self.model_config.fc_repr_dim,
                        bias=True,
                    ),
                }
            )
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.no_out_channels * self.data_size_after_conv

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.conv[0].conv_1.weight

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_repr_dim

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # Swish slope is roughly 0.5 around 0
                nn.init.kaiming_normal_(tensor=m.weight, a=0.5, mode="fan_out")

    @property
    def residual_blocks(self) -> List[int]:
        mc = self.model_config
        if not mc.layers:
            stride_w = mc.down_stride_width
            stride_h = mc.down_stride_height

            if stride_w < 2 and stride_h < 2:
                raise ValueError(
                    "At least one of the strides must be greater than 1, "
                    "when automatic residual block calculation is used"
                    "for CNN model creation. Got strides: "
                    f"down_stride_width: {stride_w} "
                    f"and down_stride_height: {stride_h}."
                )

            residual_blocks = auto_find_no_cnn_residual_blocks_needed(
                input_size_w=self.data_dimensions.width,
                kernel_w=mc.kernel_width,
                first_kernel_expansion_w=mc.first_kernel_expansion_width,
                stride_w=mc.down_stride_width,
                first_stride_expansion_w=mc.first_stride_expansion_width,
                dilation_w=mc.dilation_factor_width,
                input_size_h=self.data_dimensions.height,
                kernel_h=mc.kernel_height,
                first_kernel_expansion_h=mc.first_kernel_expansion_height,
                stride_h=mc.down_stride_height,
                first_stride_expansion_h=mc.first_stride_expansion_height,
                dilation_h=mc.dilation_factor_height,
                cutoff=mc.cutoff,
            )

            logger.debug(
                "No residual blocks specified in CL args, using input "
                "%s based on size approximation calculation.",
                residual_blocks,
            )
            return residual_blocks

        return self.model_config.layers

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = self.pos_representation(input)
        out = self.conv(out)
        out = out.view(out.shape[0], -1)

        out = self.fc(out)

        return out


def _make_conv_layers(
    residual_blocks: List[int],
    cnn_model_configuration: CNNModelConfig,
    data_dimensions: "DataDimensions",
) -> List[nn.Module]:
    """
    Used to set up the convolutional layers for the model. Based on the passed in
    residual blocks, we want to set up the actual blocks with all the relevant
    convolution parameters.

    We start with a base channel number of 2**5 == 32.

    Also inserts a self-attention layer in just before the last residual block.

    :param residual_blocks: List of ints, where each int indicates number of blocks w.
    that channel dimension.
    :param cnn_model_configuration: Experiment hyperparameters / configuration needed
    for the convolution setup.
    :return: A list of ``nn.Module`` objects to be passed to ``nn.Sequential``.
    """
    mc = cnn_model_configuration

    first_conv_channels = 2**mc.channel_exp_base * mc.first_channel_expansion

    down_stride_w = mc.down_stride_width
    first_conv_kernel_w = mc.kernel_width * mc.first_kernel_expansion_width
    first_conv_stride_w = down_stride_w * mc.first_stride_expansion_width

    first_kernel_w, first_pad_w = pytorch_utils.calc_conv_params_needed(
        input_size=data_dimensions.width,
        kernel_size=first_conv_kernel_w,
        stride=first_conv_stride_w,
        dilation=1,
    )

    down_stride_h = mc.down_stride_height
    first_conv_kernel_h = mc.kernel_height * mc.first_kernel_expansion_height
    first_conv_stride_h = mc.down_stride_height * mc.first_stride_expansion_height

    first_kernel_h, first_pad_h = pytorch_utils.calc_conv_params_needed(
        input_size=data_dimensions.height,
        kernel_size=first_conv_kernel_h,
        stride=first_conv_stride_h,
        dilation=1,
    )

    conv_blocks = [
        FirstCNNBlock(
            in_channels=data_dimensions.channels,
            out_channels=first_conv_channels,
            conv_1_kernel_h=first_kernel_h,
            conv_1_kernel_w=first_kernel_w,
            conv_1_padding_w=first_pad_w,
            conv_1_padding_h=first_pad_h,
            down_stride_w=first_conv_stride_w,
            dilation_w=1,
            dilation_h=1,
            rb_do=mc.rb_do,
        )
    ]

    for layer_arch_idx, layer_arch_layers in enumerate(residual_blocks):
        for layer in range(layer_arch_layers):
            cur_layer = _get_conv_residual_block(
                conv_blocks=conv_blocks,
                layer_arch_idx=layer_arch_idx,
                down_stride_w=down_stride_w,
                down_stride_h=down_stride_h,
                cnn_config=cnn_model_configuration,
                data_dimensions=data_dimensions,
            )

            conv_blocks.append(cur_layer)

            cur_width, cur_height = pytorch_utils.calc_size_after_conv_sequence(
                input_width=data_dimensions.width,
                input_height=data_dimensions.height,
                conv_sequence=nn.Sequential(*conv_blocks),
            )

            if cur_height * cur_width <= mc.attention_inclusion_cutoff:
                cur_attention_block = ConvAttentionBlock(
                    channels=cur_layer.out_channels,
                    width=cur_width,
                    height=cur_height,
                )
                conv_blocks.append(cur_attention_block)

    return conv_blocks


def _get_conv_residual_block(
    conv_blocks: List[nn.Module],
    layer_arch_idx: int,
    down_stride_w: int,
    down_stride_h: int,
    cnn_config: CNNModelConfig,
    data_dimensions: "DataDimensions",
) -> CNNResidualBlock:
    mc = cnn_config

    cur_conv = nn.Sequential(*conv_blocks)
    cur_width, cur_height = pytorch_utils.calc_size_after_conv_sequence(
        input_width=data_dimensions.width,
        input_height=data_dimensions.height,
        conv_sequence=cur_conv,
    )

    cur_block_number = (
        len([i for i in conv_blocks if isinstance(i, CNNResidualBlock)]) + 1
    )
    cur_dilation_factor_w = _get_cur_dilation(
        dilation_factor=mc.dilation_factor_width,
        size=cur_width,
        block_number=cur_block_number,
    )

    cur_dilation_factor_h = _get_cur_dilation(
        dilation_factor=mc.dilation_factor_height,
        size=cur_height,
        block_number=cur_block_number,
    )

    cur_kernel_w, cur_padding_w = pytorch_utils.calc_conv_params_needed(
        input_size=cur_width,
        kernel_size=mc.kernel_width,
        stride=down_stride_w,
        dilation=cur_dilation_factor_w,
    )

    cur_kernel_h, cur_padding_h = pytorch_utils.calc_conv_params_needed(
        input_size=cur_height,
        kernel_size=mc.kernel_height,
        stride=down_stride_h,
        dilation=cur_dilation_factor_h,
    )

    cur_in_channels = conv_blocks[-1].out_channels
    cur_out_channels = 2 ** (mc.channel_exp_base + layer_arch_idx)

    cur_layer = CNNResidualBlock(
        in_channels=cur_in_channels,
        out_channels=cur_out_channels,
        conv_1_kernel_w=cur_kernel_w,
        conv_1_padding_w=cur_padding_w,
        conv_1_kernel_h=cur_kernel_h,
        conv_1_padding_h=cur_padding_h,
        down_stride_w=down_stride_w,
        dilation_w=cur_dilation_factor_w,
        dilation_h=cur_dilation_factor_h,
        full_preact=True if len(conv_blocks) == 1 else False,
        rb_do=mc.rb_do,
        stochastic_depth_p=mc.stochastic_depth_p,
    )

    return cur_layer


def _get_cur_dilation(dilation_factor: int, size: int, block_number: int):
    """
    Note that block_number refers to the number of residual blocks (not first block
    or self attention).
    """

    if size == 1:
        return 1

    dilation = dilation_factor**block_number

    while dilation >= size:
        dilation = dilation // dilation_factor

    return dilation


def auto_find_no_cnn_residual_blocks_needed(
    input_size_w: int,
    kernel_w: int,
    first_kernel_expansion_w: int,
    stride_w: int,
    first_stride_expansion_w: int,
    dilation_w: int,
    input_size_h: int,
    kernel_h: int,
    first_kernel_expansion_h: int,
    stride_h: int,
    first_stride_expansion_h: int,
    dilation_h: int,
    cutoff: int,
) -> List[int]:
    """
    Used in order to calculate / set up residual blocks specifications as a list
    automatically when they are not passed in as CL args, based on the minimum
    size after the residual block convolutions.

    We have 2 residual_blocks per channel depth until we have a total of 8 blocks,
    then the rest is put in the third depth index (following resnet convention).

    That is with a base channel depth of 32, we have these depths in the list:
    [32, 64, 128, 256].

    Examples
    ------
    3 blocks --> [2, 1]
    7 blocks --> [2, 2, 2, 1]
    10 blocks --> [2, 2, 4, 2]
    """

    if (stride_w == 1 and input_size_w > cutoff) or (
        stride_h == 1 and input_size_h > cutoff
    ):
        err_dim = "width" if stride_w == 1 else "height"
        logger.warning(
            f"With stride=1, the {err_dim} size "
            f"({input_size_w if err_dim == 'width' else input_size_h}) "
            f"cannot be larger than the cutoff ({cutoff}). "
            f"This would result in an infinite loop."
            f"Will stop when no more reduction is possible,"
            f"despite not reaching the cutoff."
        )

    k_size_w_first, padding_w_first = pytorch_utils.calc_conv_params_needed(
        input_size=input_size_w,
        kernel_size=kernel_w * first_kernel_expansion_w,
        stride=stride_w * first_stride_expansion_w,
        dilation=dilation_w,
    )

    k_size_h_first, padding_h_first = pytorch_utils.calc_conv_params_needed(
        input_size=input_size_h,
        kernel_size=kernel_h * first_kernel_expansion_h,
        stride=stride_h * first_stride_expansion_h,
        dilation=dilation_h,
    )

    cur_size_w = pytorch_utils.conv_output_formula(
        input_size=input_size_w,
        kernel_size=k_size_w_first,
        stride=stride_w * first_stride_expansion_w,
        dilation=dilation_w,
        padding=padding_w_first,
    )
    cur_size_h = pytorch_utils.conv_output_formula(
        input_size=input_size_h,
        kernel_size=k_size_h_first,
        stride=stride_h * first_stride_expansion_h,
        dilation=dilation_h,
        padding=padding_h_first,
    )

    residual_blocks = [0] * 4
    while True:
        cur_kernel_w, cur_padding_w = pytorch_utils.calc_conv_params_needed(
            input_size=cur_size_w,
            kernel_size=kernel_w,
            stride=stride_w,
            dilation=dilation_w,
        )

        cur_kernel_h, cur_padding_h = pytorch_utils.calc_conv_params_needed(
            input_size=cur_size_h,
            kernel_size=kernel_h,
            stride=stride_h,
            dilation=dilation_h,
        )

        cur_size_w_next = pytorch_utils.conv_output_formula(
            input_size=cur_size_w,
            kernel_size=cur_kernel_w,
            stride=stride_w,
            dilation=dilation_w,
            padding=cur_padding_w,
        )

        cur_size_h_next = pytorch_utils.conv_output_formula(
            input_size=cur_size_h,
            kernel_size=cur_kernel_h,
            stride=stride_h,
            dilation=dilation_h,
            padding=cur_padding_h,
        )

        cannot_reduce_more = (
            cur_size_w == cur_size_w_next and cur_size_h == cur_size_h_next
        )
        w_kernel_larger = cur_kernel_w > cur_size_w
        h_kernel_larger = cur_kernel_h > cur_size_h
        kernel_too_large = w_kernel_larger or h_kernel_larger
        if (
            cur_size_w_next * cur_size_h_next < cutoff
            or cannot_reduce_more
            or kernel_too_large
        ):
            if cannot_reduce_more:
                logger.warning(
                    f"Could not reduce size more, "
                    f"despite not reaching the cutoff ({cutoff})."
                )

            break
        else:
            cur_no_blocks = sum(residual_blocks)

            if cur_no_blocks >= 8:
                residual_blocks[2] += 1
            else:
                cur_index = cur_no_blocks // 2
                residual_blocks[cur_index] += 1

            cur_size_w, cur_size_h = cur_size_w_next, cur_size_h_next

    return [i for i in residual_blocks if i != 0]


class GeneralPositionalEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        max_length: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.max_length = max_length
        self.dropout = nn.Dropout(p=dropout)

        self.embedding = torch.nn.Parameter(
            data=torch.zeros(1, self.embedding_dim, self.max_length),
            requires_grad=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.embedding
        return self.dropout(x)



from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Sequence, Literal, Tuple, Dict, Iterable

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import ensure_path_exists, get_logger
from captum.attr._utils.visualization import (
    format_classname,
    format_word_importances,
    _get_color,
)
from torchtext.vocab import Vocab

from eir.interpretation.interpretation_utils import (
    get_target_class_name,
    stratify_attributions_by_target_classes,
    plot_attributions_bar,
    get_basic_sample_attributions_to_analyse_generator,
)

if TYPE_CHECKING:
    from eir.train import Experiment
    from eir.interpretation.interpretation import SampleAttribution

logger = get_logger(name=__name__)


@dataclass()
class SequenceVisualizationDataRecord:
    sample_id: str
    token_attributions: Sequence[float]
    label_name: str
    attribution_score: float
    raw_input_tokens: Sequence[str]


def analyze_sequence_input_attributions(
    experiment: "Experiment",
    input_name: str,
    output_name: str,
    target_column_name: str,
    target_column_type: str,
    attribution_outfolder: Path,
    all_attributions: Sequence["SampleAttribution"],
    expected_target_classes_attributions: Sequence[float],
) -> None:
    exp = experiment

    output_object = exp.outputs[output_name]
    target_transformer = output_object.target_transformers[target_column_name]

    input_object = exp.inputs[input_name]
    interpretation_config = input_object.input_config.interpretation_config

    samples_to_act_analyze_gen = get_basic_sample_attributions_to_analyse_generator(
        interpretation_config=interpretation_config, all_attributions=all_attributions
    )
    vocab = exp.inputs[input_name].vocab
    viz_records = []

    for sample_attribution in samples_to_act_analyze_gen:
        sample_target_labels = sample_attribution.sample_info.target_labels

        cur_label_name = get_target_class_name(
            sample_label=sample_target_labels[output_name][target_column_name],
            target_transformer=target_transformer,
            column_type=target_column_type,
            target_column_name=target_column_name,
        )

        extracted_sample_info = extract_sample_info_for_sequence_attribution(
            sample_attribution_object=sample_attribution,
            cur_label_name=cur_label_name,
            output_name=output_name,
            target_column_name=target_column_name,
            target_column_type=target_column_type,
            input_name=input_name,
            vocab=vocab,
            expected_target_classes_attributions=expected_target_classes_attributions,
        )

        index_to_truncate = get_sequence_index_to_truncate_unknown(
            raw_inputs=extracted_sample_info.raw_inputs
        )

        truncated_sample_info = truncate_sequence_attribution_to_padding(
            sequence_attribution_sample_info=extracted_sample_info,
            truncate_start_idx=index_to_truncate,
        )
        tsi = truncated_sample_info

        if not tsi.raw_inputs:
            logger.debug(
                "Skipping sequence attribution analysis of single sample %s as it is "
                "empty after truncating unknowns.",
                sample_attribution.sample_info.ids[0],
            )
            continue

        viz_record = SequenceVisualizationDataRecord(
            sample_id=sample_attribution.sample_info.ids[0],
            token_attributions=tsi.sequence_attributions,
            label_name=tsi.sample_target_label_name,
            attribution_score=tsi.sequence_attributions.sum(),
            raw_input_tokens=tsi.raw_inputs,
        )
        viz_records.append(viz_record)

    html_string = get_sequence_html(data_records=viz_records)
    outpath = attribution_outfolder / "single_samples.html"
    ensure_path_exists(path=outpath)
    save_html(out_path=outpath, html_string=html_string)

    acts_stratified_by_target = stratify_attributions_by_target_classes(
        all_attributions=all_attributions,
        target_transformer=target_transformer,
        output_name=output_name,
        target_column=target_column_name,
        column_type=target_column_type,
    )

    for class_name, target_attributions in acts_stratified_by_target.items():
        token_importances = get_sequence_token_importance(
            attributions=target_attributions, vocab=vocab, input_name=input_name
        )
        df_token_importances = get_sequence_feature_importance_df(
            token_importances=token_importances
        )
        target_outfolder = attribution_outfolder / f"{class_name}"
        ensure_path_exists(path=target_outfolder, is_folder=True)
        plot_attributions_bar(
            df_attributions=df_token_importances,
            outpath=target_outfolder / f"token_influence_{class_name}.pdf",
            title=f"{target_column_name} – {class_name}",
        )
        df_token_importances.to_csv(
            path_or_buf=target_outfolder / f"token_influence_{class_name}.csv"
        )


@dataclass
class SequenceAttributionSampleInfo:
    sequence_attributions: np.ndarray
    raw_inputs: Sequence[str]
    expected_attr_value: float
    sample_target_label_name: str


def extract_sample_info_for_sequence_attribution(
    sample_attribution_object: "SampleAttribution",
    cur_label_name: str,
    output_name: str,
    target_column_name: str,
    target_column_type: str,
    input_name: str,
    vocab: Vocab,
    expected_target_classes_attributions: Sequence[float],
) -> SequenceAttributionSampleInfo:
    attributions = sample_attribution_object.sample_attributions[input_name]

    sample_tokens = sample_attribution_object.raw_inputs[input_name]
    raw_inputs = extract_raw_inputs_from_tokens(tokens=sample_tokens, vocab=vocab)

    cur_sample_expected_value = _parse_out_sequence_expected_value(
        sample_target_labels=sample_attribution_object.sample_info.target_labels,
        output_name=output_name,
        target_column_name=target_column_name,
        target_column_type=target_column_type,
        expected_values=expected_target_classes_attributions,
    )

    extracted_sequence_info = SequenceAttributionSampleInfo(
        sequence_attributions=attributions,
        raw_inputs=raw_inputs,
        expected_attr_value=cur_sample_expected_value,
        sample_target_label_name=cur_label_name,
    )

    return extracted_sequence_info


def extract_raw_inputs_from_tokens(tokens: torch.Tensor, vocab) -> Sequence[str]:
    raw_inputs = vocab.lookup_tokens(tokens.squeeze().tolist())
    return raw_inputs


def _parse_out_sequence_expected_value(
    sample_target_labels: Dict[str, Dict[str, torch.Tensor]],
    output_name: str,
    target_column_name: str,
    expected_values: Sequence[float],
    target_column_type: str,
) -> float:
    if target_column_type == "con":
        assert len(expected_values) == 1
        return expected_values[0]

    cur_base_values_index = sample_target_labels[output_name][target_column_name].item()
    cur_sample_expected_value = expected_values[cur_base_values_index]

    return cur_sample_expected_value


def get_sequence_index_to_truncate_unknown(
    raw_inputs: Sequence[str], padding_value: str = "<pad>"
) -> int:
    raw_inputs_reversed = raw_inputs[::-1]
    counter = 0
    for element in raw_inputs_reversed:
        if element == padding_value:
            counter += 1
        else:
            break

    index_to_truncate = len(raw_inputs) - counter

    return index_to_truncate


def truncate_sequence_attribution_to_padding(
    sequence_attribution_sample_info: SequenceAttributionSampleInfo,
    truncate_start_idx: int,
) -> SequenceAttributionSampleInfo:
    si = sequence_attribution_sample_info

    attrs_truncated, raw_inputs_truncated = _truncate_attributions_and_raw_inputs(
        attributions=si.sequence_attributions,
        raw_inputs=si.raw_inputs,
        truncate_start_idx=truncate_start_idx,
    )

    truncated_attribution = SequenceAttributionSampleInfo(
        sequence_attributions=attrs_truncated,
        raw_inputs=raw_inputs_truncated,
        expected_attr_value=si.expected_attr_value,
        sample_target_label_name=si.sample_target_label_name,
    )

    return truncated_attribution


def _truncate_attributions_and_raw_inputs(
    attributions: np.ndarray, raw_inputs: Sequence[str], truncate_start_idx: int
) -> Tuple[np.ndarray, Sequence[str]]:
    attrs_copy = copy(attributions)
    raw_inputs_copy = copy(raw_inputs)

    attrs_summed = attrs_copy.squeeze().sum(1)
    attrs_truncated = attrs_summed[:truncate_start_idx]

    raw_inputs_truncated = [i + " " for i in raw_inputs_copy][:truncate_start_idx]

    return attrs_truncated, raw_inputs_truncated


def save_html(out_path: Path, html_string: str) -> None:
    with open(out_path, "w", encoding="utf-8") as outfile:
        outfile.write(html_string)


def get_label_transformer_mapping(
    transformer, order: Literal["int-to-string", "string-to-int"]
) -> dict:
    values = transformer.classes_, transformer.transform(transformer.classes_)
    if order == "int-to-string":
        values = transformer.transform(transformer.classes_), transformer.classes_

    mapping = dict(zip(*values))

    return mapping


def get_sequence_token_importance(
    attributions: Sequence["SampleAttribution"], vocab: Vocab, input_name: str
) -> Dict[str, list[float]]:
    token_importances = defaultdict(list)

    for act in attributions:
        orig_seq_input = extract_raw_inputs_from_tokens(
            tokens=act.raw_inputs[input_name], vocab=vocab
        )
        seq_attrs = act.sample_attributions[input_name]
        seq_attrs_abs = seq_attrs.squeeze()
        assert len(seq_attrs_abs.shape) == 2

        seq_attrs_values_sum = seq_attrs_abs.sum(1).squeeze()
        assert len(seq_attrs_values_sum) == len(orig_seq_input)

        for token, attribution in zip(orig_seq_input, seq_attrs_values_sum):
            token_importances[token].append(attribution)

    return token_importances


def get_sequence_feature_importance_df(
    token_importances: Dict[str, list[float]]
) -> pd.DataFrame:
    df = pd.concat({k: pd.Series(v) for k, v in token_importances.items()})
    df = df.reset_index(level=0).reset_index(drop=True)
    df.columns = ["Input", "Attribution"]

    return df


def get_sequence_html(
    data_records: Iterable[SequenceVisualizationDataRecord], legend: bool = True
) -> str:
    dom = [
        "<table width: 100%; border: 2px solid black; border-collapse: collapse;>",
        '<table class="table" style="text-align: center;">',
    ]
    rows = [
        "<tr><th>ID</th>"
        "<th>True Label</th>"
        "<th>Attribution Score</th>"
        "<th>Token Importance</th>"
    ]
    for record in data_records:
        rows.append(
            "".join(
                [
                    "<tr>",
                    format_classname(record.sample_id),
                    format_classname(record.label_name),
                    format_classname("{0:.2f}".format(record.attribution_score)),
                    format_word_importances(
                        record.raw_input_tokens, record.token_attributions
                    ),
                    "<tr>",
                ]
            )
        )

    if legend:
        dom.append(
            '<div style="border-top: 1px solid; margin-top: 5px; \
            padding-top: 5px; display: inline-block">'
        )
        dom.append("<b>Legend: </b>")

        for value, label in zip([-1, 0, 1], ["Negative", "Neutral", "Positive"]):
            dom.append(
                '<span style="display: inline-block; width: 10px; height: 10px; \
                border: 1px solid; background-color: \
                {value}"></span> {label}  '.format(
                    value=_get_color(value), label=label
                )
            )
        dom.append("</div>")

    dom.append("".join(rows))
    dom.append("</table>")
    html = "".join(dom)

    return html
